/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

/* Example: ex_mroot.cpp
   -------------------------------------------------------------------
   Several ways to use an O2scl solver to solve a simple function
*/

#include <cmath>
#include <o2scl/test_mgr.h>
#include <o2scl/mm_funct.h>
#include <o2scl/gsl_mroot_hybrids.h>
#include <o2scl/cern_mroot.h>

using namespace std;
using namespace o2scl;

int gfn(size_t nv, const ovector_base &x, 
	ovector_base &y, void *&pa) {
  y[0]=sin(x[1]-0.2);
  y[1]=sin(x[0]-0.25);
  return 0;
}

class cl {

public:

  // Store the number of function and derivative evaluations
  int nf, nd;

  int mfn(size_t nv, const ovector_base &x, ovector_base &y, 
	  void *&pa) {
    y[0]=sin(x[1]-0.2);
    y[1]=sin(x[0]-0.25);
    nf++;
    return 0;
  }

  int operator()(size_t nv, const ovector_base &x, 
		 ovector_base &y, void *&pa) {
    y[0]=sin(x[1]-0.2);
    y[1]=sin(x[0]-0.25);
    nf++;
    return 0;
  }

  int mfnd(size_t nv, ovector_base &x, ovector_base &y,
	   omatrix_base &j, void *&pa) {
    j[0][0]=0.0;
    j[0][1]=cos(x[1]-0.2);
    j[1][0]=cos(x[0]-0.25);
    j[1][1]=0.0;
    nd++;
    return 0;
  }

  int mfna(size_t nv, const double x[2], double y[2], void *&pa) {
    y[0]=sin(x[1]-0.2);
    y[1]=sin(x[0]-0.25);
    return 0;
  }

  int mfnad(size_t nv, double x[], double y[], double j[2][2], void *&pa) {
    j[0][0]=0.0;
    j[0][1]=cos(x[1]-0.2);
    j[1][0]=cos(x[0]-0.25);
    j[1][1]=0.0;
    return 0;
  }

};

int main(void) {
  cl acl;
  ovector x(2);
  double xa[2];
  int i;
  void *vp=NULL;
  size_t tmp;
  int r1, r2, r3;
  bool done;
  test_mgr t;

  t.set_output_level(1);
  
  /*
    Using a member function with \ref ovector objects
  */
  mm_funct_mfptr<cl,void *,ovector_base> f1(&acl,&cl::mfn);
  gsl_mroot_hybrids<void *> cr1;
    
  x[0]=0.5;
  x[1]=0.5;
  acl.nf=0;
  int ret1=cr1.msolve(2,x,vp,f1);
  cout << "GSL solver (numerical Jacobian): " << endl;
  cout << "Return value: " << ret1 << endl;
  cout << "Number of iterations: " << cr1.last_ntrial << endl;
  cout << "Number of function evaluations: " << acl.nf << endl;
  cout << endl;
  t.test_rel(x[0],0.25,1.0e-6,"1a");
  t.test_rel(x[1],0.2,1.0e-6,"1b");

  /*
    Using the CERNLIB solver
  */
  cern_mroot<void *> cr2;
    
  x[0]=0.5;
  x[1]=0.5;
  acl.nf=0;
  int ret2=cr2.msolve(2,x,vp,f1);
  cout << "CERNLIB solver (numerical Jacobian): " << endl;
  cout << "Return value: " << ret2 << endl;
  cout << "INFO parameter: " << cr2.get_info() << endl;
  cout << "Number of function evaluations: " << acl.nf << endl;
  cout << endl;
  t.test_rel(x[0],0.25,1.0e-6,"2a");
  t.test_rel(x[1],0.2,1.0e-6,"2b");

  /*
    Using a member function with \ref ovector objects, but
    using the GSL-like interface with set() and iterate().
  */
  gsl_mroot_hybrids<void *> cr3;

  x[0]=0.5;
  x[1]=0.5;
  cr3.allocate(2);
  cr3.set(2,x,f1,vp);
  done=false;
  do {
    r3=cr3.iterate();
    double resid=fabs(cr3.f[0])+fabs(cr3.f[1]);
    if (resid<cr3.tolf || r3>0) done=true;
  } while (done==false);
  t.test_rel(cr3.x[0],0.25,1.0e-6,"3a");
  t.test_rel(cr3.x[1],0.2,1.0e-6,"3b");
  cr3.free();

  /*
    Now instead of using the automatic Jacobian, using 
    a user-specified Jacobian.
  */
  jac_funct_mfptr<cl,void *,ovector_base,omatrix_base> j4(&acl,&cl::mfnd);
    
  x[0]=0.5;
  x[1]=0.5;
  acl.nf=0;
  acl.nd=0;
  int ret4=cr1.msolve_de(2,x,vp,f1,j4);
  cout << "GSL solver (analytic Jacobian): " << endl;
  cout << "Return value: " << ret4 << endl;
  cout << "Number of iterations: " << cr1.last_ntrial << endl;
  cout << "Number of function evaluations: " << acl.nf << endl;
  cout << "Number of Jacobian evaluations: " << acl.nd << endl;
  cout << endl;
  t.test_rel(x[0],0.25,1.0e-6,"4a");
  t.test_rel(x[1],0.2,1.0e-6,"4b");

  /*
    Using a user-specified Jacobian and the GSL-like interface
  */
  gsl_mroot_hybrids<void *> cr5;

  x[0]=0.5;
  x[1]=0.5;
  cr5.allocate(2);
  cr5.set_de(2,x,f1,j4,vp);
  done=false;
  do {
    r3=cr5.iterate();
    double resid=fabs(cr5.f[0])+fabs(cr5.f[1]);
    if (resid<cr5.tolf || r3>0) done=true;
  } while (done==false);
  t.test_rel(cr5.x[0],0.25,1.0e-6,"5a");
  t.test_rel(cr5.x[1],0.2,1.0e-6,"5b");
  cr5.free();

  /*
    Using C-style arrays instead of ovector objects
  */
  mm_vfunct_mfptr<cl,void *,2> f6(&acl,&cl::mfna);
  gsl_mroot_hybrids<void *,mm_vfunct_mfptr<cl,void *,2>,double[2],
    double[2],array_alloc<double[2]> > cr6;
    
  xa[0]=0.5;
  xa[1]=0.5;
  cr6.msolve(2,xa,vp,f6);
  t.test_rel(xa[0],0.25,1.0e-6,"6a");
  t.test_rel(xa[1],0.2,1.0e-6,"6b");
  
  /*
    Using the CERNLIB solver with C-style arrays instead of ovector objects
  */
  cern_mroot<void *,mm_vfunct_mfptr<cl,void *,2>,double[2],
    double[2],array_alloc<double[2]> > cr7;
    
  xa[0]=0.5;
  xa[1]=0.5;
  cr7.msolve(2,xa,vp,f6);
  t.test_rel(xa[0],0.25,1.0e-6,"7a");
  t.test_rel(xa[1],0.2,1.0e-6,"7b");
  
  /*
    Using C-style arrays with a user-specified Jacobian
  */
  jac_vfunct_mfptr<cl,void *,2> j8(&acl,&cl::mfnad);
  gsl_mroot_hybrids<void *,mm_vfunct_mfptr<cl,void *,2>,double[2],
    double[2],array_alloc<double[2]>,double[2][2],double[2][2],
    array_2d_alloc<double[2][2]>,jac_vfunct<void *,2> > cr8;
    
  xa[0]=0.5;
  xa[1]=0.5;
  cr8.msolve_de(2,xa,vp,f6,j8);
  t.test_rel(xa[0],0.25,1.0e-6,"8a");
  t.test_rel(xa[1],0.2,1.0e-6,"8b");
  
  /*
    Using a class with an operator(). Note that there can be only one
    operator() function in each class.
  */
  gsl_mroot_hybrids<void *,cl,ovector_base> cr9;

  x[0]=0.5;
  x[1]=0.5;
  cr9.msolve(2,x,vp,acl);
  t.test_rel(x[0],0.25,1.0e-6,"9a");
  t.test_rel(x[1],0.2,1.0e-6,"9b");

  /*
    Using a function pointer to a global function.
  */
  typedef int (*gfnt)(size_t, const ovector_base &, ovector_base &, 
		      void *&);
  gsl_mroot_hybrids<void *,gfnt,ovector_base> cr10;
  gfnt f10=&gfn;

  x[0]=0.5;
  x[1]=0.5;
  cr10.msolve(2,x,vp,f10);
  t.test_rel(x[0],0.25,1.0e-6,"10a");
  t.test_rel(x[1],0.2,1.0e-6,"10b");

  t.report();
  return 0;
}
// End of example
