/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/string_conv.h>
#include <o2scl/err_hnd.h>

using namespace std;
using namespace o2scl;

std::string o2scl::btos(bool b) {
  if (b) return "1";
  return "0";
}

bool o2scl::has_minus_sign(double *x) {
  gsl_ieee_double_rep r;
  gsl_ieee_double_to_rep(x, &r);
  if (r.sign==1) return true;
  return false;
}

std::string o2scl::double_to_ieee_string(const double *x) {
    
  gsl_ieee_double_rep r;
    
  gsl_ieee_double_to_rep(x, &r);

  std::string s;
    
  char sign=' ';
  if (r.sign==1) sign='-';
    
  switch (r.type)
    {
    case GSL_IEEE_TYPE_NAN:
      s="NaN";
      break ;
    case GSL_IEEE_TYPE_INF:
      s=sign;
      s+="Inf";
      break ;
    case GSL_IEEE_TYPE_NORMAL:
      s=sign;
      s+="1.";
      s+=r.mantissa;
      s+="*2^";
      s+=itos(r.exponent);
      break ;
    case GSL_IEEE_TYPE_DENORMAL:
      s=sign;
      s+="0.";
      s+=r.mantissa;
      s+="*2^";
      s+=itos(r.exponent+1);
      break ;
    case GSL_IEEE_TYPE_ZERO:
      s=sign;
      s+="0";
      break ;
    default:
      s="[non-standard IEEE float]";
    }
  return s;
}
  
string o2scl::itos(int x) {
  string ret;
  ostringstream strout;
  strout.setf(ios::scientific);
  
  if (strout << x) {
    return strout.str();
  }
  
  O2SCL_ERR("Conversion from integer to string failed in itos().",
	    gsl_einval);
  return "";
}

string o2scl::ptos(void *p) {
  string ret;
  ostringstream strout;
  strout.setf(ios::scientific);

  if (strout << p) {
    return strout.str();
  }
  
  O2SCL_ERR("Conversion from pointer to string failed in ptos().",
	    gsl_einval);
  return "";
}

std::string o2scl::dtos(double x, std::ostream &format) {
  string ret;
  ostringstream strout;
  
  strout.precision(format.precision());
  strout.fill(format.fill());
  strout.flags(format.flags());

  if (strout << x) {
    return strout.str();
  } 
  
  O2SCL_ERR("Conversion from double to string failed in dtos(double,ostream).",
	  gsl_einval);
  return "";
}

size_t o2scl::size_of_exponent(double x) {
  string ret;
  ostringstream strout;
  
  strout.setf(ios::scientific);

  if (strout << x) {
    ret=strout.str();
    if (ret.length()==0 || ret.find('e')==std::string::npos) {
      O2SCL_ERR("No exponent found in size_of_exponent().",gsl_einval);
      return 0;
    }
    return ret.length()-ret.find('e')-2;
  }

  O2SCL_ERR("Failed to convert to string in size_of_exponent().",gsl_einval);
  return 0;
}
  
string o2scl::dtos(double x, int prec, bool auto_prec) {
  string ret;
  ostringstream strout;

  if (prec!=0) {
    if (!auto_prec) strout.setf(ios::scientific);
    strout.precision(prec);
  }

  if (strout << x) {
    return strout.str();
  }
  
  O2SCL_ERR2("Conversion from double to string failed in ",
	     "dtos(double,int,bool).",gsl_einval);
  return "";
}

int o2scl::stoi(string s) {
  int ret;
  istringstream ins(s);
  if (ins >> ret) {
    return ret;
  }
  O2SCL_ERR("Conversion from integer to string failed in stoi().",
	  gsl_einval);
  return 0;
}

bool o2scl::stob(string s) {
  bool ret;
  // we read into a string stream to remove initial whitespace
  istringstream ins(s);
  string s2;
  if (ins >> s2) {
    if (s2.length()>0) {
      if (s2[0]=='t' || s2[0]=='T') return true;
      if (s2[0]>'0' && s2[0]<='9') return true;
    }
  }
  return false;
}

double o2scl::stod(string s) {
  double ret;
  istringstream ins(s);
  if (ins >> ret) {
    return ret;
  } 
  O2SCL_ERR("Conversion from string to double failed in o2scl::stod().",
	  gsl_efailed);
  return 0.0;
}

