/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_STRING_CONV_H
#define O2SCL_STRING_CONV_H
/** \file string_conv.h
    \brief Various string conversion functions
*/

#include <iostream>
#include <cmath>
#include <string>
#include <fstream>
#include <sstream>
#include <o2scl/lib_settings.h>
#include <gsl/gsl_ieee_utils.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Convert a pointer to a string 

      This uses an \c ostringstream to convert a pointer to a string
      and is architecture-dependent.
  */
  std::string ptos(void *p);

  /** \brief Convert an integer to a string 
   */
  std::string itos(int x);

  /** 
      \brief Convert a boolean value to a string 

      This returns \c "1" for true and \c "0" for false.
   */
  std::string btos(bool b);

  /** \brief Convert a double to a string 

      If \c auto_prec is false, then the number is converted to 
      a string in the <tt>ios::scientific</tt> mode, otherwise,
      neither the scientific or fixed mode flags are set and the
      number is converted to a string in "automatic" mode.
   */
  std::string dtos(double x, int prec=6, bool auto_prec=false);

  /** \brief Returns the number of characters required to display the 
      exponent of \c x in scientific mode
      
      This returns 2 or 3, depending on whether or not the 
      absolute magnitude of the exponent is greater than 100. It
      uses <tt>stringstream</tt> to convert the number to a string
      and counts the number of characters directly.
   */
  size_t size_of_exponent(double x);

  /** 
      \brief Convert a double to a string using a specified format
  */
  std::string dtos(double x, std::ostream &format);

  /** 
      \brief Convert a string to an integer 
      
      If this function fails it will call O2SCL_ERR() and return zero.
  */
  int stoi(std::string s);

  /** 
      \brief Convert a string to a boolean value
      
      This returns true only if the string has at least one character
      and the first non-whitespace character is either \c t, \c T, or
      one of the numbers 1 through 9.
      
      This function never fails (it just returns false for an empty
      string).
  */
  bool stob(std::string s);

  /** 
      \brief Convert a string to a double 

      If this function fails it will call O2SCL_ERR() and return zero.
  */
  double stod(std::string s);

  /** 
      \brief Convert a double to a string containing IEEE representation
      
      Modeled after the GSL function \c gsl_ieee_fprintf_double(), but
      converts to a \c string instead of a \c FILE *.
  */
  std::string double_to_ieee_string(const double *x);
  
  /** 
      \brief Find out if the number pointed to by \c x has a minus sign
      
      This function returns true if the number pointed to by \c x has
      a minus sign using the GSL IEEE functions. It is useful, for
      example, in distinguishing "-0.0" from "+0.0".
  */
  bool has_minus_sign(double *x);

#ifndef DOXYGENP
}
#endif

#endif

