/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_TABLE_H
#define O2SCL_TABLE_H

#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
#include <sstream>
#include <o2scl/misc.h>
#include <o2scl/err_hnd.h>
#include <o2scl/multi_funct.h>
#include <o2scl/collection.h>
#include <o2scl/search_vec.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/smart_interp.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Data \table class
      
      \b Summary \n 

      A class to contain and manipulate several equally-sized columns
      of data. The purpose of this class is to provide a structure
      which allows one to refer to the columns using a name
      represented by a string. Thus for a table object named \c t with
      3 columns (named "colx", "coly" and "colz") and three rows, one
      could do the following:
      \include table_doc1.cpp
      Note that the rows are numbered starting with 0 instead of 
      starting with 1.
      To output all the rows of entire column, one can use
      \include table_doc2.cpp
      To output all the columns of an entire row (in the following
      example it is the second row), labeled by their column name, one
      can use:
      \include table_doc3.cpp

      Methods are provided for interpolating columns, sorting 
      columns, finding data points, and several other manipulations
      of the data. 

      \b Column \b size \n 

      The columns grow automatically (similar to the STL \<vector\>)
      in reponse to an attempt to call set() for a row that does not
      presently exist or in a call to line_of_data() when the table is
      already full. However, this forces memory rearrangments that are
      O(R*C). If the user has a good estimate of the number of rows
      beforehand, it is best to either specify this in the
      constructor, or in an explicit call to inc_maxlines().

      <B> Lookup, differentiation, integration, and interpolation </b> \n

      Lookup, differentiation, integration, and interpolation are
      automatically implemented using splines from the class \ref
      smart_interp_vec. A caching mechanism is implemented so that
      successive interpolations, derivative evaluations or
      integrations over the same two columns are fast.

      <B> Sorting </b>\n

      The columns are automatically sorted by name for speed, the
      results can be accessed by table::get_sorted_name(i). Individual
      columns can be sorted (\ref sort_column() ), or the entire table
      can be sorted by one column (\ref sort_table() ).

      <B> Allowable column names</b> \n
      
      In general, column names may be of any form as long as they
      don't contain whitespace, e.g. \c 123".#@$%xy~ is a legitmate
      column name. The \o2x extension to this class named
      <tt>table_fp</tt>, however requires that the column name contain
      only letters, numbers, and underscores and also that the column
      name not begin with a digit.
      
      \comment
      <B> User-owned columns </b> \n
      \endcomment

      <B> Data representation </b> \n

      Each individual column is just an ovector_view (or any
      descendant of an ovector_view) The columns can be referred to in
      one of two ways:
      - A numerical index from 0 to C-1 (where C is the number of
      columns). For example, data can be accessed through \ref
      table::get() and \ref table::set(size_t c, size_t r,
      double val), or the overloaded [] operator, table[c][r]. 
      - A name of the column which is a string with no whitespace.
      For example, data can be accessed with table::get(string cname,
      int r) and table::set(string cname, int r, double val).
      
      The columns are organized in a both a \<map\> and a \<vector\>
      structure so that finding a column by its index ( string
      table::get_column_name(int index), and double
      *table::get_column(int index) ) takes only constant time, and
      finding a column by its name ( int \ref lookup_column() and
      double * \ref table::get_column() ) is O(log(C)). Insertion of a
      column ( \ref new_column() ) is O(log(C)), but deletion ( \ref
      delete_column() ) is O(C). Adding a row of data can be either
      O(1) or O(C), but row insertion and deletion is slow, since the
      all of the columns must be shifted accordingly.
      
      \comment
      Ownership of any column may be changed at any time, but care
      must be taken to ensure that memory allocation errors do not
      occur. These errors should not occur when no columns are 
      owned by the user.
      \endcomment

      Because of the structure, this class is not suitable for the
      matrix manipulation. The classes \ref omatrix and \ref umatrix
      are better used for that purpose.

      <B> Thread-safety </b> \n

      Generally, the member functions are thread-safe in the sense
      that one would expect: it is always permitted to have many
      threads accessing many tables, but care must be taken if many
      threads are accessing only one table. Simple get() and set()
      functions are thread-safe (unless a set() function forces memory
      rearrangement), while insertion and deletion operations are not.
      Only the <tt>const</tt> version of the interpolation routines
      are thread-safe

      \b I/O \b and \b command-line \b manipulation \n

      When data from an object of type \ref table is output to 
      a file through the \ref collection class, the table
      can be manipulated on the command-line through
      the acol utility.

      There is an example for the usage of this class given
      in <tt>examples/ex_table.cpp</tt>.

      \future Rewrite the \ref table::create_array() and 
      \ref table::insert_data()
      functions for generic vector type

      \future Be more restrictive about allowable column names?
      
      \future Re-implement user-owned columns, and handle automatic
      resizing appropriately.
      
      \comment

      10/16/07: This issue may be unimportant, as it might
      be better to just move to a template based approach with
      a user-specified vector type. The interpolation is now flexible
      enough to handle different types. Might check to ensure sorting
      works with other types. 

      10/9/08: Actually, the native vector
      types are very useful here, and I'm thinking now that the
      vector types should not be templated. If there is a user-owned
      column, then no lines should ever be added beyond the length
      of the user-owned column. For adding user-owned columns 
      into a table which already contains data, it should fail
      if the user-owned column is smaller than nlines and succeed
      otherwise. If after adding a user-owned column maxlines is
      greater than the user-owned column size, then maxlines should
      be decreases accordingly.

      1/30/08: User-owned columns have been removed for now,
      which avoids a lot of complications

      \endcomment

      \future Return the empty column in the operator[] functions
      as is done for the get_column() functions.

      \future A "delete rows" method to delete a range of several rows

      \future The present structure, \c
      std::map<std::string,col,string_comp> atree and \c
      std::vector<aiter> alist; could be replaced with \c
      std::vector<col> list and \c std::map<std::string,int> tree
      where the map just stores the index of the the column in the
      list
  */
  class table {
  public:
  
    /** \brief Create a new table with space for nlines<=cmaxlines.
     */
    table(int cmaxlines=0);

    virtual ~table();

    /// Copy constructor
    table(const table &t);

    /// Copy constructor
    table &operator=(const table &t);

    // --------------------------------------------------------
    /** \name Basic get and set methods */
    //@{
    /** \brief Set row \c row of column named \c col to value \c val - 
	O(log(C)).

	This function adds the column \c col if it does not
	already exist and adds rows using inc_maxlines() and
	set_nlines() to create at least \c (row+1) rows if they do not
	already exist.
     */
    int set(std::string col, size_t row, double val);

    /** \brief Set row \c row of column number \c icol to value \c val - O(1).
     */
    int set(size_t icol, size_t row, double val);
  
    /** \brief Get value from row \c row of column named \c col - O(log(C)).
     */
    double get(std::string col, size_t row) const;

    /** \brief Get value from row \c row of column number \c icol - O(1).
     */
    double get(size_t icol, size_t row) const;

    /** \brief Return the number of columns
     */
    size_t get_ncolumns() const {return atree.size();};
    
    /** \brief Return the number of lines
     */
    size_t get_nlines() const {return nlines;};

    /** 
	\brief Set the number of lines

	This function is stingy about increasing the table memory space
	and will only increase it enough to fit \c il lines, which
	is useful if you have columns not owned by the table.
     */
    int set_nlines(size_t il);

    /** 
	\brief Set the number of lines, increasing the size more 
	agressively

	This function is like set_nlines(), but doubles the maximum
	column size if an increase in the maximum size is required 
	instead of simply making enough room for the current number of 
	lines. This function is used internally by \ref set() to 
	ensure that the cost of setting lines in sequence is linear
	and not quadratic. 
    */
    int set_nlines_auto(size_t il);

    /** \brief Return the maximum number of lines
     */
    int get_maxlines() {return maxlines;};

    /** \brief Returns a reference to the column named \c col - O(log(C))
     */
    ovector_base &get_column(std::string col);
    
    /** \brief Returns a reference to the column named \c col - O(log(C))
     */
    const ovector_base &get_column(std::string col) const;

    /** 
	\brief Returns the column of index \c icol - O(1) (const version)

	This does not do any sort of bounds checking and is quite
	fast.

        Note that several of the methods require reallocation of
	memory and refereces previously returned by this function will
	be incorrect.
    */
    const ovector_base &operator[] (size_t icol) const {
      return (*alist[icol]->second.dat);
    }

    /** 
	\brief Returns the column of index \c icol - O(1).

	This does not do any sort of bounds checking and is 
	quite fast. 

        Note that several of the methods require reallocation of
	memory and refereces previously returned by this function will
	be incorrect.
    */
    ovector_base &operator[] (size_t icol) {
      return (*alist[icol]->second.dat);
    }

    /** 
	\brief Returns the column named \c scol - O(log(C)) (const version)

	No error checking is performed.

        Note that several of the methods require reallocation of
	memory and refereces previously returned by this function will
	be incorrect.
    */
    const ovector_base &operator[](std::string scol) const {
      aciter it=atree.find(scol);
      return *(it->second.dat);
    }

    /** 
	\brief Returns the column named \c scol - O(log(C))

	No error checking is performed.

        Note that several of the methods require reallocation of
	memory and refereces previously returned by this function will
	be incorrect.
    */
    ovector_base &operator[](std::string scol) {
      aiter it=atree.find(scol);
      return *(it->second.dat);
    }

    /** \brief Returns a copy of the row with value \c val 
	in column \c col - O(R*C) */
    int get_row(std::string col, double val, ovector &row) const;
    
    /** \brief Returns a copy of row number \c irow - O(C)
     */
    int get_row(size_t irow, ovector &row) const;
    //@}
  
    // --------------------------------------------------------
    /** \name Column manipulation */
    //@{

    /** \brief Returns the name of column \c col - O(1).
     */
    std::string get_column_name(size_t col) const;

    /** \brief Returns the name of column \c col in sorted order - O(1).
     */
    std::string get_sorted_name(size_t col);

    /** \brief Add a new column owned by the table - O(log(C)).
     */
    int new_column(std::string name);

#ifdef O2SCL_NEVER_DEFINED
    /** \brief Add a new column owned by the user - O(log(C)).
	
	This function does not modify the number of lines of data in 
	the table. 

	\todo We've got to figure out what to do if ldat is too small.
	If it's smaller than nlines, obviously we should just 
	fail, but what if it's size is between nlines and maxlines?
    */
    int new_column(std::string name, ovector_view *ldat);
#endif

    /** \brief Return true if \c scol is a column in the current table

        This function does not call the error handler if the 
	column is not found, but just silently returns false.
     */
    bool is_column(std::string scol);

    /** 
	\brief Find the index for column named \c name - O(log(C)).

	If the column is not present, this does not call the error
	handler, but quietly sets \c ix to zero and returns \ref
	gsl_enotfound.
     */
    int lookup_column(std::string name, int &ix) const;

#ifdef O2SCL_NEVER_DEFINED
    /** 
	\brief Rename column named \c olds to \c news - O(C). 

	This is slow since we have to delete the column and re-insert
	it. This process in turn mangles all of the 
	iterators in the list.
    */
    int rename_column(std::string olds, std::string news);
#endif

    /** \brief Make a new column named \c dest equal to \c src - O(log(C)*R).
     */
    int copy_column(std::string src, std::string dest);

    /// Create (using \c new) a generic array from column \c col
    double *create_array(std::string col) const;

    /** \brief Initialize all values of column named \c scol to \c val - 
	O(log(C)*R). 

	Note that this does not initialize elements beyond
	nlines so that if the number of rows is increased afterwards, the
	new rows will have uninitialized values.
    */
    int init_column(std::string scol, double val);

#ifdef O2SCL_NEVER_DEFINED
    /** 
	\brief Modify ownership - O(log(C)). 
	
	\warning columns allocated using malloc() should never be
	owned by the table object since it uses delete instead of
	free().
    */
    int ch_owner(std::string name, bool ow);
#endif

    /** 
	\brief Get a gsl_vector from column \c name - O(log(C)). 
    */
    const gsl_vector *get_gsl_vector(std::string name) const;
  
    /** 
	\brief Insert a column from a separate table, interpolating
	it into a new column
	
	Given a pair of columns ( \c src_index, \c src_col ) in a
	separate table (\c source), this creates a new column in the
	present table named \c src_col which interpolates \c loc_index
	into \c src_index.  The interpolation objects from the \c
	source table will be used. If there is already a column in the
	present table named \c src_col, then this will fail.

	If there is an error in the interpolation for any particular
	row, then the value of \c src_col in that row will be set to
	zero.
    */
    int add_col_from_table(std::string loc_index, table &source,
			   std::string src_index, std::string src_col,
			   std::string dest_col="");
    //@}
  
    // --------------------------------------------------------
    /** \name Row maninpulation and data input */
    //@{

    /** \brief Insert a row before row \c n 

	Acceptable values for \c n are between 0 and
	<tt>get_nlines()</tt> inclusive, with the maximum value
	denoting the addition of a row after the last row presently in
	the table.
     */
    int new_row(size_t n);

    /** \brief Copy the data in row \c src  to row \c dest 
     */
    int copy_row(size_t src, size_t dest);

    /** \brief Insert a row of data before row \c n 
     */
    int insert_data(size_t n, size_t nv, double *v);

    /** \brief Insert a row of data before row \c n 
     */
    int insert_data(size_t n, size_t nv, double **v);

    /** 
	\brief Read a new set of names from \c newheads
    */
    int line_of_names(std::string newheads);

    /** \brief Read a line of data from an array
     */
    template<class vec_t> int line_of_data(size_t nv, const vec_t &v) {
      if (maxlines==0) inc_maxlines(5);
      if (nlines>=maxlines) inc_maxlines(maxlines);
      
      if (intp_set) {
	intp_set=false;
	delete si;
      }
      
      if (nlines<maxlines && nv<=(atree.size())) {
	
	for(size_t i=0;i<nv;i++) {
	  (*this)[i][nlines]=v[i];
	}
	nlines++;
	
	return 0;
      }
      
      O2SCL_ERR("Not enough lines or columns in line_of_data().",gsl_einval);
      return gsl_einval;
    }
    //@}

    // --------------------------------------------------------
    /** \name Lookup and search methods */
    //@{

    /** 
	\brief Look for a value in an ordered column

	O(log(C)*log(R))

	This uses the function search_vec::ordered_lookup(), which
	offers caching and assumes the vector is monotonic. If you
	don't have monotonic data, you can still use the
	table::lookup() function, which is more general.
     */
    size_t ordered_lookup(std::string col, double val);

    /// Exhaustively search column \c col for the value \c val - O(log(C)*R).
    size_t lookup(std::string col, double val) const;

    /// Search column \c col for the value \c val and return value in \c col2
    double lookup_val(std::string col, double val, std::string col2) const;

    /// Exhaustively search column \c col for the value \c val - O(log(C)*R).
    size_t lookup(int col, double val) const;

    /** \brief Exhaustively search column \c col for many occurences 
	of \c val - O(log(C)*R).
    */
    size_t mlookup(std::string col, double val, std::vector<double> &results,
		   double threshold=0.0) const;
    //@}

    // --------------------------------------------------------
    /** \name Interpolation, differentiation, and integration, max, and min */
    //@{

    /// Set the base interpolation objects
    int set_interp(base_interp_mgr<ovector_const_view> &bi1,
		   base_interp_mgr<ovector_const_subvector> &bi2) {
      bim1=&bi1;
      bim2=&bi2;
      return 0;
    }

    /** 
	\brief Interpolate \c x0 from \c sx into \c sy

	O(log(C)*log(R)) but can be as bad as O(log(C)*R) if 
	the relevant columns are not well ordered.
    */
    double interp(std::string sx, double x0, std::string sy);

    /** 
	\brief Interpolate \c x0 from \c sx into \c sy

	O(log(C)*log(R)) but can be as bad as O(log(C)*R) if 
	the relevant columns are not well ordered.
    */
    double interp_const(std::string sx, double x0, std::string sy) const;
    
    /** 
	\brief Interpolate \c x0 from \c ix into \c iy

	O(log(R)) but can be as bad as O(R) if 
	the relevant columns are not well ordered.
     */
    double interp(size_t ix, double x0, size_t iy);

    /** 
	\brief Interpolate \c x0 from \c ix into \c iy

	O(log(R)) but can be as bad as O(R) if 
	the relevant columns are not well ordered.
     */
    double interp_const(size_t ix, double x0, size_t iy) const;

    /** \brief Make a new column \c yp which is the derivative y'(x) - 
	O(log(C)*R).
     */
    int deriv(std::string x, std::string y, std::string yp);

    /** 
	\brief The first derivative of the function sy(sx) at sx=x0.

	O(log(C)*log(R)) but can be as bad as O(log(C)*R) if 
	the relevant columns are not well ordered.
    */
    double deriv(std::string sx, double x0, std::string sy);

    /** 
	\brief The first derivative of the function sy(sx) at sx=x0.

	O(log(C)*log(R)) but can be as bad as O(log(C)*R) if 
	the relevant columns are not well ordered.
    */
    double deriv_const(std::string sx, double x0, std::string sy) const;

    /** 
	\brief The first derivative of the function iy(ix) at ix=x0.
      
	O(log(R)) but can be as bad as O(R) if 
	the relevant columns are not well ordered.
    */
    double deriv(size_t ix, double x0, size_t iy);

    /** 
	\brief The first derivative of the function iy(ix) at ix=x0.
      
	O(log(R)) but can be as bad as O(R) if 
	the relevant columns are not well ordered.
    */
    double deriv_const(size_t ix, double x0, size_t iy) const;

    /** 
	\brief Make a new column \c yp which is 
	\f$ y^{\prime \prime}(x) \f$ - O(log(C)*R).
     */
    int deriv2(std::string x, std::string y, std::string yp);

    /** 
	\brief The second derivative of the function sy(sx) at sx=x0.

	O(log(C)*log(R)) but can be as bad as O(log(C)*R) if 
	the relevant columns are not well ordered.
    */
    double deriv2(std::string sx, double x0, std::string sy);

    /** 
	\brief The second derivative of the function sy(sx) at sx=x0.

	O(log(C)*log(R)) but can be as bad as O(log(C)*R) if 
	the relevant columns are not well ordered.
    */
    double deriv2_const(std::string sx, double x0, std::string sy) const;

    /** 
	\brief The second derivative of the function iy(ix) at ix=x0.

	O(log(R)) but can be as bad as O(R) if 
	the relevant columns are not well ordered.
    */
    double deriv2(size_t ix, double x0, size_t iy);

    /** 
	\brief The second derivative of the function iy(ix) at ix=x0.

	O(log(R)) but can be as bad as O(R) if 
	the relevant columns are not well ordered.
    */
    double deriv2_const(size_t ix, double x0, size_t iy) const;

    /** 
	\brief The integral of the function sy(sx) from sx=x1 to sx=x2.
	
	O(log(C)*log(R)) but can be as bad as O(log(C)*R) if 
	the relevant columns are not well ordered.
    */
    double integ(std::string sx, double x1, double x2, std::string sy);
  
    /** 
	\brief The integral of the function sy(sx) from sx=x1 to sx=x2.
	
	O(log(C)*log(R)) but can be as bad as O(log(C)*R) if 
	the relevant columns are not well ordered.
    */
    double integ_const(std::string sx, double x1, double x2, 
		       std::string sy) const;
  
    /** 
	\brief The integral of the function iy(ix) from ix=x1 to ix=x2.

	O(log(R)) but can be as bad as O(R) if 
	the relevant columns are not well ordered.
    */
    double integ(size_t ix, double x1, double x2, size_t iy);

    /** 
	\brief The integral of the function iy(ix) from ix=x1 to ix=x2.

	O(log(R)) but can be as bad as O(R) if 
	the relevant columns are not well ordered.
    */
    double integ_const(size_t ix, double x1, double x2, size_t iy) const;

    /** 
	\brief The integral of the function iy(ix) 

	O(log(R)) but can be as bad as O(R) if the relevant columns
	are not well ordered.
    */
    int integ(std::string x, std::string y, std::string ynew);

    /** \brief Return column maximum. Makes no assumptions about 
	ordering - O(R).
     */
    double max(std::string col) const;

    /** \brief Return column minimum. Makes no assumptions about 
	ordering - O(R).
     */
    double min(std::string col) const;
    //@}

    // --------------------------------------------------------
    /** \name Subtable method */
    //@{

    /** 
	\brief Make a subtable

	Uses the columns specified in \c list from the row \c top
	to the row of index \c bottom to generate a new table
	which is a copy of part of the original.

	\comment
	The 'linked' variable was taken out since we're removing
	user-owned columns for now

	If \c linked is false
	the the data will be independent from the original table.
	\endcomment
    */
    table *subtable(std::string list, size_t top, size_t bottom) const;
    //@}
  
    // --------------------------------------------------------
    /** \name Add space */
    //@{

    /** \brief Manually increase the maximum number of lines
     */
    int inc_maxlines(size_t llines);
    //@}

    // --------------------------------------------------------
    /** \name Delete methods */
    //@{

    /** \brief Delete column named \c scol - O(C). 
    
	This is slow because the iterators in \ref alist are mangled
	and we have to call \ref reset_list() to get them back.
    */
    virtual int delete_column(std::string scol);

    /** \brief Delete the row with the entry closest to
	the value \c val in column \c scol - O(R*C)
     */
    int delete_row(std::string scol, double val);

    /** \brief Delete the row of index \c irow - O(R*C)
     */
    int delete_row(size_t irow);
    //@}

    // --------------------------------------------------------
    /** \name Clear methods */
    //@{

    /** \brief Zero the data entries but keep the column names 
	and nlines fixed
    */
    void zero_table();

    /** \brief Clear the table and the column names
     */
    void clear_table();

    /** \brief Remove all of the data by setting the number
	of lines to zero

        This leaves the column names intact and does not remove
	the constants.
     */
    void clear_data() {
      nlines=0;   
      if (intp_set==true) {
	delete si; 
	intp_set=false;
      }
      return;
    };
    //@}

    // --------------------------------------------------------
    /** \name Sorting methods */
    //@{

    /** \brief Sort the entire table by the column \c scol
     */
    int sort_table(std::string scol);

    /** \brief Individually sort the column \c scol
     */
    int sort_column(std::string scol);
    //@}

    // --------------------------------------------------------
    /** \name Summary method */
    //@{
    /** \brief Output a summary of the information stored

	Outputs the number of constants, the number of columns,
	a list of the column names, and the number of lines of
	data.
    */
    virtual int summary(std::ostream *out, int ncol=79) const;
    //@}

    /// \name Constant manipulation
    //@{
    /// Add a constant, or if the constant already exists, change its value
    virtual int add_constant(std::string name, double val) {
      if (constants.find(name)!=constants.end()) {
	constants.find(name)->second=val;
	return 0;
      }
      constants.insert(make_pair(name,val));
      return 0;
    }

    /// Add a constant
    virtual int set_constant(std::string name, double val,
			     bool err_on_notfound=true) {
      if (constants.find(name)!=constants.end()) {
	constants.find(name)->second=val;
	return 0;
      }
      if (err_on_notfound) {
	O2SCL_ERR_RET("No constant with specified name in set_constant().",
		      gsl_einval);
      }
      return gsl_enotfound;
    }

    /// Get a constant
    virtual double get_constant(std::string name) const {
      return constants.find(name)->second;
    }

    /// Get the number of constants
    virtual size_t get_nconsts() const {
      return constants.size();
    }

    /// Get a constant by index
    virtual int get_constant(size_t ix, std::string &name, double &val) const {
      if (ix<constants.size()) {
	std::map<std::string,double>::const_iterator cit=constants.begin();
	for(size_t i=0;i<ix;i++) cit++;
	name=cit->first;
	val=cit->second;
	return 0;
      }
      O2SCL_ERR_RET("Index too large in table::get_constant().",gsl_eindex);
    }

    /// Remove a constant
    virtual int remove_constant(std::string name) {
      constants.erase(name);
      return 0;
    }
    //@}
    
    /// \name Miscellaneous methods
    //@{
    /// Clear the current table and read from a generic data file
    int read_generic(std::istream &fin);

    /** \brief Return 0 if the tree and list are properly synchronized
     */
    int check_synchro() const;

    /** \brief Get ownership - O(log(C)). 
     */
    bool get_owner(std::string name) const;

    //@}

    /// Default interpolation manager
    def_interp_mgr<ovector_const_view,cspline_interp> dim1;

    /// Default interpolation manager
    def_interp_mgr<ovector_const_subvector,cspline_interp> dim2;

    /// Return the type, \c "table".
    virtual const char *type() { return "table"; }

    /// True if the interpolation type has been set
    bool intp_set;

  protected:

#ifndef DOXYGEN_INTERNAL
    
    friend class io_tlate<table>;

    /** \brief The list of constants 
     */
    std::map<std::string,double> constants;

    /** \brief Set the elements of alist with the appropriate 
	iterators from atree - O(C).

	Generally, the end-user shouldn't need this method. It is 
	only used in delete_column() to rearrange the list when
	a column is deleted from the tree.
    */
    int reset_list();

    /** 
       \brief Ensure a variable name does not match a function or contain 
       non-alphanumeric characters
    */
    int make_fp_varname(std::string &s);

    /// Make sure a name is unique
    int make_unique_name(std::string &col, std::vector<std::string> &cnames);

    /// Column structure for \ref table [protected]
    typedef struct col_s {
      /// Pointer to column
      ovector_base *dat;
      /// Owner of column
      bool owner;
      /// Column index
      int index;
    } col;
    
    /// \name Iterator types
    //@{
    typedef std::map<std::string,col,string_comp>::iterator aiter;
    typedef std::map<std::string,col,string_comp>::const_iterator aciter;
    typedef std::vector<aiter>::iterator aviter;
    //@}
  
    /// \name Actual data
    //@{
    /// The size of allocated memory
    size_t maxlines;
    /// The size of presently used memory
    size_t nlines;
    /// The tree of columns
    std::map<std::string,col,string_comp> atree;
    /// The list of tree iterators
    std::vector<aiter> alist;
    //@}
  
    /// \name Column manipulation methods
    //@{
    /// Return the iterator for a column
    aiter get_iterator(std::string lname);
    /// Return the column structure for a column
    col *get_col_struct(std::string lname);
    /// Return the beginning of the column tree
    aiter begin() {return atree.begin();};
    /// Return the end of the column tree
    aiter end() {return atree.end();};
    //@}

    /// An empty vector for get_column()
    ovector empty_col;

    /// The sorting function
    static int sortd_comp(const void *a, const void *b);

    /// A structure for sorting in \ref table [protected]
    typedef struct sortd_s {
      /// Value to sort
      double val;
      /// Sorted index
      int indx;
    } sortd;
  
    /// \name Interpolation
    //@{
    /// The interpolation object
    sm_interp_vec *si;

    /// A pointer to the interpolation manager
    base_interp_mgr<ovector_const_view> *bim1;
    /// A pointer to the subvector interpolation manager
    base_interp_mgr<ovector_const_subvector> *bim2;

    /// The vector-searching object
    search_vec<ovector> se;
    
    /// The last x-column interpolated
    std::string intp_colx;

    /// The last y-column interpolated
    std::string intp_coly;

    //@}

#endif

  };

  /// Input specialization
  template<> int io_tlate<table>::input
    (cinput *co, o2scl::in_file_format *ins, table *ta);
  /// Output specialization
  template<> int io_tlate<table>::output
    (coutput *co, o2scl::out_file_format *outs, 
     table *at);
  /// \fn Type specialization
  template<> const char *io_tlate<table>::type();
  
  typedef io_tlate<table> table_io_type;
  
#ifndef DOXYGENP
}
#endif

#endif
