/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#include <o2scl/table3d.h>
#include <o2scl/test_mgr.h>
#include <o2scl/base_ioc.h>
#include <o2scl/user_io.h>

using namespace std;
using namespace o2scl;

int main(void) {

  cout.setf(ios::scientific);

  test_mgr t;
  t.set_output_level(1);

  table3d at;
  at.set_size(2,4);
  at.set_size(3,5);

  // Set the grid size
  size_t nx1, ny1;
  at.get_size(nx1,ny1);
  cout << "nx: " << nx1 << " ny: " << ny1 << endl;
    
  // Add some slices
  int r1=at.new_slice("z1");
  t.test_gen(r1==0,"slice 1");
  int r2=at.new_slice("z2");
  t.test_gen(r2==0,"slice 2");

  // Try to reset the size when slices have been added.
  // Show that the size is unchanged.
  def_err_hnd.set_mode(0);
  at.set_size(2,6);
  def_err_hnd.set_mode(2);
  at.get_size(nx1,ny1);
  t.test_gen(nx1==3,"size 1");
  t.test_gen(ny1==5,"size 2");
  cout << "nx: " << nx1 << " ny: " << ny1 << endl;
  cout << endl;

  // Add the data
  for(size_t i=0;i<nx1;i++) {
    for(size_t j=0;j<ny1;j++) {
      int er1=at.set(i,j,"z1",sqrt(((double)(i+j))));
      t.test_gen(er1==0,"set 1");
      int er2=at.set(i,j,"z2",sqrt(2.0*((double)(i+j))));
      t.test_gen(er2==0,"set 2");
    }
  }
    
  // Definite the grid
  double xv[3]={0,1,2};
  double yv[5]={0,1,2,3,4};
  at.set_xy<double *>("x",3,xv,"y",5,yv);

  // Output the table3d object
  cout.precision(5);

  cout << " z1:        ";
  for(size_t i=0;i<5;i++) {
    cout << yv[i] << " ";
  }
  cout << endl;
  for(size_t i=0;i<nx1;i++) {
    cout << xv[i] << " ";
    for(size_t j=0;j<ny1;j++) {
      cout << at.get(i,j,"z1") << " ";
    }
    cout << endl;
  }
  cout << endl;
    
  cout << " z2:        ";
  for(size_t i=0;i<5;i++) {
    cout << yv[i] << " ";
  }
  cout << endl;
  for(size_t i=0;i<nx1;i++) {
    cout << xv[i] << " ";
    for(size_t j=0;j<ny1;j++) {
      cout << at.get(i,j,"z2") << " ";
    }
    cout << endl;
  }
  cout << endl;

  // Test the interpolation
  cout << "1.5,0.5,z2: " << at.interp(1.5,0.5,"z2") << endl;
  cout << "1.9,0.75,z2: " << at.interp(1.9,0.75,"z2") << endl;
  cout << "2.0,1.0,z2: " << at.interp(2.0,1.0,"z2") << endl;
  cout << "2.1,1.25,z2: " << at.interp(2.1,1.25,"z2") << endl;
  cout << "2.5,1.5,z2: " << at.interp(2.5,1.5,"z2") << endl;
  cout << endl;

  // Print out the summary
  at.summary(&cout);
  cout << endl;

  // Use the I/O facilities to output the object
  base_ioc bio;
  table3d_io_type t3dio;
  io_type_info iti;
  iti.type_summary(&cout);
  
  o2scl_output_text("table3d.dat",&at,"test");
  //tof=new text_out_file("table3d.dat");
  //co.out_one(tof,"table3d","test",&at);
  //delete tof;

  // Read the object from a file
  table3d *at2;
  //std::string name;
  //int sz, sz2;
  //void *vp;

  o2scl_input_text("table3d.dat",at2);

  //tif=new text_in_file("table3d.dat");
  //co.in_one(tif,"table3d",name,vp,sz,sz2);
  //at2=(table3d *)vp;
  //delete tif;

  // Print out the copy read in from the file
  at2->summary(&cout);
  cout << endl;

  cout << " z1:        ";
  for(size_t i=0;i<5;i++) {
    cout << yv[i] << " ";
  }
  cout << endl;
  for(size_t i=0;i<3;i++) {
    cout << xv[i] << " ";
    for(size_t j=0;j<5;j++) {
      cout << at2->get(i,j,"z1") << " ";
    }
    cout << endl;
  }
  cout << endl;

  cout << " z2:        ";
  for(size_t i=0;i<5;i++) {
    cout << yv[i] << " ";
  }
  cout << endl;
  for(size_t i=0;i<3;i++) {
    cout << xv[i] << " ";
    for(size_t j=0;j<5;j++) {
      cout << at2->get(i,j,"z2") << " ";
    }
    cout << endl;
  }
  cout << endl;

  // Test extract functions
  table tx, ty;
  at2->extract_x(1.0,tx);
  at2->extract_y(2.0,ty);

  for(size_t i=0;i<tx.get_ncolumns();i++) {
    cout.width(12);
    cout << tx.get_column_name(i) << " ";
  }
  cout << endl;
  for(size_t j=0;j<tx.get_nlines();j++) {
    for(size_t i=0;i<tx.get_ncolumns();i++) {
      cout.width(12);
      cout << tx.get(i,j) << " ";
    }
    cout << endl;
  }
  cout << endl;

  for(size_t i=0;i<ty.get_ncolumns();i++) {
    cout.width(12);
    cout << ty.get_column_name(i) << " ";
  }
  cout << endl;
  for(size_t j=0;j<ty.get_nlines();j++) {
    for(size_t i=0;i<ty.get_ncolumns();i++) {
      cout.width(12);
      cout << ty.get(i,j) << " ";
    }
    cout << endl;
  }
  cout << endl;

  t.report();

  return 0;
}


