/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_USER_IO_H
#define O2SCL_USER_IO_H

#include <o2scl/collection.h>

/** \file user_io.h
    \brief Support functions for I/O
    
    \todo Do input for array and 2d array objects, finish output
    functions. Make a macro for I/O for doubles, ints and other
    objects w/o a type() function.
*/

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Input one object from a file
    
      This requires a default copy constructor and a type() function.
  
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.
  */
  template<class obj_t> 
    int o2scl_input(in_file_format *ins, obj_t *&obj,
		    std::string &name, size_t &sz, size_t &sz2, 
		    bool verbose=false) {
    
    obj_t odum;
    std::string stype=odum.type();

    // Read the file
    collection co;
    int ret=co.fin(ins);
    if (ret!=0) {
      O2SCL_ERR_RET("Read of file failed in o2scl_input().",gsl_efailed);
    }
    
    // Look for the object in the file
    collection::type_iterator it=co.begin(stype);
    if (it==co.end(stype)) {
      O2SCL_ERR_RET((((std::string)"Couldn't find object of type '")+stype+
		     "' in collection in function o2scl_input(in_"+
		     "file_format,obj_t,string,size_t,size_t,bool).").c_str(),
		    gsl_efailed);
    }
    // Double check that the pointer is nonzero
    if (it->data==0) {
      O2SCL_ERR_RET("Null pointer in function o2scl_input().",
		  gsl_efailed);
    }
    // Check that sizes are correct
    if (it->size==0 || it->size2==0) {
      O2SCL_ERR_RET("Size of object incorrect in o2scl_input().",
		  gsl_efailed);
    }

    // Assign name, pointer, and size
    name=it.name();
    sz=it->size;
    sz2=it->size2;
    obj=(obj_t *)it->data;
    
    // Give ownership to the user
    co.disown(name);
    
    // Added on 4/20/06...good idea? (6/6/08 - I think this is good,
    // but this should all be done in the collection destructor anyway)
    co.clear();

    return ret;
  }

  /** 
      \brief Input one object from a file
    
      This requires a default copy constructor and a type() function.
  
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.
  */
  template<class obj_t> 
    int o2scl_input(in_file_format *ins, obj_t *&obj,
		    std::string &name, size_t &sz, bool verbose=false) {
    
    obj_t odum;
    std::string stype=odum.type();

    // Read the file
    collection co;
    int ret=co.fin(ins);
    if (ret!=0) {
      O2SCL_ERR_RET("Read of file failed in o2scl_input().",gsl_efailed);
    }
    
    // Look for the object in the file
    collection::type_iterator it=co.begin(stype);
    if (it==co.end(stype)) {
      O2SCL_ERR_RET((((std::string)"Couldn't find object of type '")+stype+
		     "' in collection in function o2scl_input().").c_str(),
		    gsl_efailed);
    }
    // Double check that the pointer is nonzero
    if (it->data==0) {
      O2SCL_ERR_RET("Null pointer in function o2scl_input().",
		  gsl_efailed);
    }
    // Check that sizes are correct
    if (it->size==0 || it->size2>0) {
      O2SCL_ERR_RET("Size of object incorrect in o2scl_input().",
		  gsl_efailed);
    }

    // Assign name, pointer, and size
    name=it.name();
    sz=it->size;
    obj=(obj_t *)it->data;
    
    // Give ownership to the user
    co.disown(name);
    
    // Added on 4/20/06...good idea? (6/6/08 - I think this is good,
    // but this should all be done in the collection destructor anyway)
    co.clear();

    return ret;
  }

  /** 
      \brief Input one object from a file
    
      This requires a default copy constructor and a type() function.
  
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.
  */
  template<class obj_t> 
    int o2scl_input(in_file_format *ins, obj_t *&obj,
		    std::string &name, bool verbose=false) {
    
    obj_t odum;
    std::string stype=odum.type();

    // Read the file
    collection co;
    int ret=co.fin(ins);
    if (ret!=0) {
      O2SCL_ERR_RET("Read of file failed in o2scl_input().",gsl_efailed);
    }
    
    // Look for the object in the file
    collection::type_iterator it=co.begin(stype);
    if (it==co.end(stype)) {
      O2SCL_ERR_RET((((std::string)"Couldn't find object of type '")+stype+
		     "' in collection in function o2scl_input("+
		     "in_file_format,obj_t,string,bool).").c_str(),
		    gsl_efailed);
    }
    // Double check that the pointer is nonzero
    if (it->data==0) {
      O2SCL_ERR_RET("Null pointer in function o2scl_input().",
		  gsl_efailed);
    }
    // Check that sizes are correct
    if (it->size>0 || it->size2>0) {
      O2SCL_ERR_RET("Size of object larger than expected in o2scl_input().",
		  gsl_efailed);
    }

    // Assign name and pointer
    name=it.name();
    obj=(obj_t *)it->data;
    
    // Give ownership to the user
    co.disown(name);
    
    // Added on 4/20/06...good idea? (6/6/08 - I think this is good,
    // but this should all be done in the collection destructor anyway)
    co.clear();

    return ret;
  }

  /** 
      \brief Input one object from a file
      
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.
  */
  template<class obj_t> 
    int o2scl_input_text(std::string fname, obj_t *&obj,
			 std::string &name, size_t &sz, size_t &sz2,
			 bool verbose=false) {
    text_in_file *tif=new text_in_file(fname);
    int ret=o2scl_input(tif,obj,name,sz,sz2,verbose);
    delete tif;
    return ret;
  }

  /** 
      \brief Input one object from a file
      
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.
  */
  template<class obj_t> 
    int o2scl_input_text(std::string fname, obj_t *&obj,
			 std::string &name, size_t &sz, bool verbose=false) {
    text_in_file *tif=new text_in_file(fname);
    int ret=o2scl_input(tif,obj,name,sz,verbose);
    delete tif;
    return ret;
  }

  /** 
      \brief Input one object from a file
      
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.
  */
  template<class obj_t> 
    int o2scl_input_text(std::string fname, obj_t *&obj,
			 std::string &name, bool verbose=false) {
    text_in_file *tif=new text_in_file(fname);
    int ret=o2scl_input(tif,obj,name,verbose);
    delete tif;
    return ret;
  }

  /** 
      \brief Input one object from a file
    
      This requires a default copy constructor and a type() function.
  
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.
  */
  template<class obj_t> 
    int o2scl_input_name(in_file_format *ins, obj_t *&obj,
			 std::string name, bool verbose=false) {
    
    obj_t odum;
    std::string stype=odum.type();

    // Read the file
    collection co;
    int ret=co.fin(ins);
    if (ret!=0) {
      O2SCL_ERR_RET("Read failed in o2scl_input_name().",gsl_efailed);
    }
    
    // Look for the object in the file
    collection::type_iterator it=co.begin(stype);
    while(name!=it.name() && it!=co.end(stype)) it++;
    if (it==co.end(stype)) {
      O2SCL_ERR_RET("Couldn't find that type in function o2scl_input_name().",
		  gsl_efailed);
    }
    if (it->data==0) {
      O2SCL_ERR_RET("Null pointer in function o2scl_input_name().",
		    gsl_esanity);
    }
    
    // Assign name and pointer
    obj=(obj_t *)it->data;
    
    // Check that sizes are correct
    ret=0;
    if (it->size>0 || it->size2>0) {
      O2SCL_ERR("Size of object larger than expected in o2scl_input_name().",
		gsl_efailed);
      ret=gsl_efailed;
    }

    // Give ownership to the user
    co.disown(name);
    
    // Added on 4/20/06...good idea? (6/6/08 - I think this is good,
    // but this should all be done in the collection destructor anyway)
    co.clear();

    return ret;
  }

  /** 
      \brief Input one object from a file
      
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.
  */
  template<class obj_t> 
    int o2scl_input_name_text(std::string fname, obj_t *&obj,
			      std::string &name, bool verbose=false) {
    text_in_file *tif=new text_in_file(fname);
    int ret=o2scl_input_name(tif,obj,name,verbose);
    delete tif;
    return ret;
  }

  /** 
      \brief Input one object from a file
    
      This requires a default copy constructor and a type() function.
  
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.
  */
  template<class obj_t> 
    int o2scl_input(in_file_format *ins, obj_t *&obj,
		    bool verbose=false) {
    
    obj_t odum;
    std::string stype=odum.type();

    // Read the file
    collection co;
    int ret=co.fin(ins);
    if (ret!=0) {
      O2SCL_ERR_RET("Read failed in o2scl_input().",gsl_efailed);
    }
    
    // Look for the object in the file
    collection::type_iterator it=co.begin(stype);
    if (it==co.end(stype)) {
      O2SCL_ERR_RET("Couldn't find that type in function o2scl_input().",
		  gsl_efailed);
    }
    if (it->data==0) {
      O2SCL_ERR_RET("Null pointer in function o2scl_input().",
		  gsl_esanity);
    }

    // Assign name and pointer
    std::string name=it.name();
    obj=(obj_t *)it->data;
    
    // Check that sizes are correct
    ret=0;
    if (it->size>0 || it->size2>0) {
      O2SCL_ERR("Size of object larger than expected in o2scl_input().",
	      gsl_efailed);
      ret=gsl_efailed;
    }

    // Give ownership to the user
    co.disown(name);
    
    // Added on 4/20/06...good idea? (6/6/08 - I think this is good,
    // but this should all be done in the collection destructor anyway)
    co.clear();

    return ret;
  }

  /** 
      \brief Input one object from a file
      
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.
  */
  template<class obj_t> 
    int o2scl_input_text(std::string fname, obj_t *&obj,
			 bool verbose=false) {
    text_in_file *tif=new text_in_file(fname);
    int ret=o2scl_input(tif,obj,verbose);
    delete tif;
    return ret;
  }

  /** 
      \brief Input one object from a file
    
      This requires a default copy constructor and a type() function.
  
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.

      \todo Do input for array and 2d array objects
  */
  template<class obj_t> 
    int o2scl_output(out_file_format *outs, obj_t *obj,
		     std::string name, size_t &sz, size_t &sz2,
		     bool verbose=false) {
    
    obj_t odum;
    std::string stype=odum.type();
    
    collection co;
    void *vp=obj;
    
    return co.out_one(outs,stype,name,vp,sz,sz2);
  }

  /** 
      \brief Input one object from a file
    
      This requires a default copy constructor and a type() function.
  
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.

      \todo Do input for array and 2d array objects
  */
  template<class obj_t> 
    int o2scl_output(out_file_format *outs, obj_t *obj,
		     std::string name, size_t &sz, bool verbose=false) {
    
    obj_t odum;
    std::string stype=odum.type();

    collection co;
    void *vp=obj;
    int isz;
    int ret=co.out_one(outs,stype,name,vp,isz);
    sz=(size_t)isz;
    return ret;
  }

  /** 
      \brief Input one object from a file
    
      This requires a default copy constructor and a type() function.
  
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.

      \todo Do input for array and 2d array objects
  */
  template<class obj_t> 
    int o2scl_output(out_file_format *outs, obj_t *obj,
		     std::string name, bool verbose=false) {
    
    obj_t odum;
    std::string stype=odum.type();
    
    collection co;
    void *vp=obj;
    return co.out_one(outs,stype,name,vp);
  }

  /** 
      \brief Input one object from a file
      
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.
  */
  template<class obj_t> 
    int o2scl_output_text(std::string fname, obj_t *obj,
			  std::string name, size_t sz,
			  size_t sz2, bool verbose=false) {
    text_out_file *tof=new text_out_file(fname);
    int ret=o2scl_output(tof,obj,name,sz,sz2,verbose);
    delete tof;
    return ret;
  }

  /** 
      \brief Input one object from a file
      
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.
  */
  template<class obj_t> 
    int o2scl_output_text(std::string fname, obj_t *obj,
			  std::string name, size_t sz,
			  bool verbose=false) {
    text_out_file *tof=new text_out_file(fname);
    int ret=o2scl_output(tof,obj,name,sz,verbose);
    delete tof;
    return ret;
  }

  /** 
      \brief Input one object from a file
      
      This does not disturb any objects in the collection. The
      pointer specified does not need to be in the collection and is
      not added to the collection.
  */
  template<class obj_t> 
    int o2scl_output_text(std::string fname, obj_t *obj,
			  std::string name, bool verbose=false) {
    text_out_file *tof=new text_out_file(fname);
    int ret=o2scl_output(tof,obj,name,verbose);
    delete tof;
    return ret;
  }


#ifndef DOXYGENP  
}
#endif

#endif
