/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/bps_eos.h>
#include <o2scl/user_io.h>
#include <o2scl/test_mgr.h>
#include <o2scl/table_units.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;
using namespace o2scl_fm;

const double xmev_kg=1.782661731e-30;
const double mev_per_fm3_msun_km3=xmev_kg/gsl_mks::solar_mass*1.0e54;

int main(void) {
  test_mgr t;
  t.set_output_level(1);
  def_err_hnd.set_mode(2);

  cout.setf(ios::scientific);

  thermo th;
  bps_eos be;
  int Z, A;
  double conv=mev_per_fm3_msun_km3;
  double cf=(gsl_cgs::mass_neutron+gsl_cgs::mass_proton)/2.0*1.0e39;
  double barn, ed_bps, pr_bps, pr;

  int sz,sz2;
  table_units *bps;
  string fname=lib_settings.get_data_dir()+"bps.eos", tname;
  
  table_units_io_type table_units_io;
  base_ioc bio;

  o2scl_input_text(fname,bps);
  
  // Cubic spline interpolation doesn't do very well here, so we use
  // linear interpolation
  def_interp_mgr<ovector_const_view,linear_interp> i1;
  def_interp_mgr<ovector_const_subvector,linear_interp> i2;
  bps->set_interp(i1,i2);

  cout << "Default mass formula, using calc_pressure()." << endl;
  cout << "eden         pres         rho          A    Z   ";
  cout << "ed_err        pr_err" << endl;
  for(pr=8.0e-10/conv/hc_mev_fm;pr>=1.0e-16/conv/hc_mev_fm;pr/=2.0) {
    th.pr=pr;
    be.calc_pressure(th,barn,Z,A);
    // Output energy and pressure in MeV/fm^3
    cout << th.ed*hc_mev_fm << " " << th.pr*hc_mev_fm
	 << " " << barn*cf << " ";
    cout.width(3);
    cout << A << " ";
    cout.width(3);
    cout << Z << " ";
    
    // Convert from g/cm^3 and dyn/cm^2 to MeV/fm^3
    ed_bps=bps->interp("nb",barn*1.0e39,"rho")/
      mev_cgs*pow(gsl_cgs::speed_of_light,2.0)/1.0e39;
    pr_bps=bps->interp("nb",barn*1.0e39,"P")/
      mev_cgs/1.0e39;

    cout.setf(ios::showpos);
    cout << (ed_bps-th.ed*hc_mev_fm)/ed_bps << " " 
	 << (pr_bps-th.pr*hc_mev_fm)/pr_bps << endl;
    cout.unsetf(ios::showpos);

    t.test_rel((ed_bps-th.ed*hc_mev_fm)/ed_bps,0.0,2.0e-3,"Energy");
    t.test_rel((pr_bps-th.pr*hc_mev_fm)/pr_bps,0.0,1.5e-1,"Pressure");
  }
  cout << endl;

  cout << "Default mass formula, using calc_density()." << endl;
  cout << "eden         pres         rho          A    Z   ";
  cout << "ed_err        pr_err" << endl;
  for(barn=7.0e11/cf;barn>=7.0e6/cf;barn/=2.0) {
    be.calc_density(barn,th,Z,A);
    
    ed_bps=bps->interp("nb",barn*1.0e39,"rho")/
      mev_cgs*pow(gsl_cgs::speed_of_light,2.0)/1.0e39;
    pr_bps=bps->interp("nb",barn*1.0e39,"P")/
      mev_cgs/1.0e39;
    
    cout << th.ed*hc_mev_fm*conv << " " << th.pr*hc_mev_fm*conv 
	 << " " << barn*cf << " ";
    cout.width(3);
    cout << A << " ";
    cout.width(3);
    cout << Z << " ";
    cout.setf(ios::showpos);
    cout << (ed_bps-th.ed*hc_mev_fm)/ed_bps << " " 
	 << (pr_bps-th.pr*hc_mev_fm)/pr_bps << endl;
    cout.unsetf(ios::showpos);
    
    t.test_rel((ed_bps-th.ed*hc_mev_fm)/ed_bps,0.0,2.0e-3,"Energy");
    t.test_rel((pr_bps-th.pr*hc_mev_fm)/pr_bps,0.0,1.5e-1,"Pressure");
  }
  cout << endl;

  cout << "Using Moller et al. mass formula, and calc_density()." << endl;
  cout << "eden         pres         rho          A    Z   ";
  cout << "ed_err        pr_err" << endl;
  mnmsk_mass mm;
  be.set_mass_formula(mm);
  for(barn=7.0e11/cf;barn>=7.0e6/cf;barn/=2.0) {
    be.calc_density(barn,th,Z,A);
    
    ed_bps=bps->interp("nb",barn*1.0e39,"rho")/
      mev_cgs*pow(gsl_cgs::speed_of_light,2.0)/1.0e39;
    pr_bps=bps->interp("nb",barn*1.0e39,"P")/
      mev_cgs/1.0e39;
    
    cout << th.ed*hc_mev_fm*conv << " " << th.pr*hc_mev_fm*conv 
	 << " " << barn*cf << " ";
    cout.width(3);
    cout << A << " ";
    cout.width(3);
    cout << Z << " ";
    cout.setf(ios::showpos);
    cout << (ed_bps-th.ed*hc_mev_fm)/ed_bps << " " 
	 << (pr_bps-th.pr*hc_mev_fm)/pr_bps << endl;
    cout.unsetf(ios::showpos);

    t.test_rel((ed_bps-th.ed*hc_mev_fm)/ed_bps,0.0,2.0e-3,"Energy");
    t.test_rel((pr_bps-th.pr*hc_mev_fm)/pr_bps,0.0,1.5e-1,"Pressure");
  }
  
  t.report();
  return 0;
}

