/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_TOV_SOLVE_H
#define O2SCL_TOV_SOLVE_H

#include <o2scl/tov_eos.h>
#include <o2scl/smart_interp.h>
#include <o2scl/table_units.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Class to solve the Tolman-Oppenheimer-Volkov equations

      Integrate Tolman-Oppenheimer-Volkov (TOV) equations

      The present code, as demonstrated in the tests, gives the
      correct central pressure and energy density of the analytical
      solution by Buchdahl to within less than 1 \part in \f$ 10^5 \f$.

      The TOV equations (i.e. Einstein's equations for a 
      static spherically symmetric object) are
      \f[
      \frac{d m}{d r} = 4 \pi r^2 \varepsilon
      \f]
      \f[
      \frac{d P}{d r} = - \frac{G \varepsilon m}{r^2}
      \left( 1+\frac{P}{\varepsilon}\right)
      \left( 1+\frac{4 \pi P r^3}{m} \right)
      \left( 1-\frac{2 G m}{r}\right)^{-1}
      \f]
      where \f$r\f$ is the radial coordinate, \f$m(r)\f$ is the mass
      enclosed within a radius \f$r\f$, and \f$\varepsilon(r)\f$ and
      \f$P(r)\f$ are the energy density and pressure at \f$r\f$, and
      \f$G\f$ is the gravitational constant. The boundary conditions
      are \f$m(r=0)=0\f$ the condition \f$P(r=R)=0\f$ for some fixed
      radius \f$R\f$. These boundary conditions give a series of
      solutions to the TOV equations as a function of the radius,
      although they do not necessarily have a solution for all radii.

      The gravitational mass is given by
      \f[
      M_G = \int_0^R 4 \pi r^2 \varepsilon d r
      \f]
      while the baryonic mass is given by
      \f[
      M_B = \int_0^R 4 \pi r^2 n_B m_B 
      \left(1-\frac{G m}{r}\right)^{-1/2} d r
      \f]
      where \f$n_B(r)\f$ is the baryon number density at radius
      \f$r\f$ and \f$m_B\f$ is the mass of one baryon.

      The gravitational potential, \f$\Phi(r)\f$ can be determined from
      \f[
      \frac{d \Phi}{d r} = - \frac{1}{\varepsilon}
      \frac{ d P}{d r} \left(1+\frac{P}{\varepsilon}\right)^{-1}
      \f]

      The proper boundary condition for the gravitational potential
      is 
      \f[
      \Phi(r=R) = \frac{1}{2} \ln \left(1-\frac{2 G M}{R} \right)
      \f]

      The equation of state may be changed at any time.
      
      Note that the pressure in the low-density eos is not 
      strictly increasing! (see at P=4.3e-10)

      The surface gravity is computed as
      \f[
      g = \frac{G M }{R^2}\left(1-\frac{2 G M}{R}\right)^{-1/2}
      \f]
      which is given in inverse kilometers and the redshift is
      \f[
      z = \left(1-\frac{2 G M}{R}\right)^{-1/2} - 1
      \f]
      which is unitless.

      <b>Screen output</b>
      - \c verbose=0 - Nothing (even if an error occurs).
      - \c verbose=1 - Some basic information and any errors or
      warnings.
      - \c verbose=2 - For each profile computation, information
      at every kilometer.
      - \c verbose=3 - Profile information every 20 grid points. A 
      keypress is required after each profile.
      
      \note The function star_fun() returns <tt>gsl_efailed</tt>
      without calling the error handler in the case that the solver
      can recover gracefully from, for example, a negative pressure.
      
      \todo
      - The error handler is called in tov_solve()::derivs for 
      pressures less than the minimum even in normal circumstances.
      This should be fixed, so that errors are more rare
      - baryon mass doesn't work for fixed() (This may be fixed. We should
      make sure it's tested.)
      - Combine maxoutsize and kmax?
      - Document column naming issues
      - Document surface gravity and redshift

      \future
      - Turn numbers in max() function into variables

      \comment
      To solve the calling of the error handler, I need to rewrite 
      the adaptive stepper to simply stop if the base stepper makes
      a step which is too large. That demands creating a new version
      of gsl_astep::evolve_apply(). The question is how to communicate
      the information from the tov derivs() function to the new version
      of gsl_astep::evolve_apply(). An easy way would be to just 
      create a new return value which indicates the step should be 
      shortened, and just stop if the step becomes too short. 
      Alternatively, instead of modifying evolve_apply(), we could just
      ensure the adaptive stepper bails out correctly if the 
      stepper returns a nonzero value and then have the caller (starfun)
      take care of it. This would probably be better. We might have 
      to make sure that gsl_astep does not set yout if the step
      fails. (3/27/09 - It looks like this might be fixed now??)
      \endcomment

  */
  class tov_solve {
  public:

    /// \name Basic properties
    //@{
    /// mass
    double mass;
    /// radius
    double rad;
    /// baryonic mass
    double bmass;
    /// gravitational potential
    double gpot;
    //@}

    /// \name Solution parameters
    /**
       These parameters can be changed at any time.
    */
    //@{
    /// Use general relativistic version (default true)
    bool generel;
    /** \brief calculate the gravitational potential and the enclosed 
	baryon mass (default false)
    */
    bool calcgpot;
    /// smallest allowed radial stepsize (default 1.0e-4)
    double hmin;
    /// largest allowed radial stepsize (default 0.05)
    double hmax;
    /// initial radial stepsize (default 4.0e-3)
    double hstart;
    /// control for output (default 1)
    int verbose;
    /// maximum radius for integration in km (default 60)
    double maxradius;
    //@}

    /// \name Mass versus radius parameters

    //@{
    /// Beginning pressure (default 7.0e-7)
    double prbegin;
    /// Ending pressure (default 8.0e-3)
    double prend;
    /// Increment for pressure (default 1.1)
    double princ;
    /// Use 'princ' as a multiplier, not an additive increment (default true)
    bool logmode;
    /** \brief Guess for central pressure in solar masses per km3 
	(default \f$ 5.2 \times 10^{-5} \f$)

	This guess is used in the function fixed().
    */
    double prguess;
    //@}

    /// \name Fixed mass parameters

    //@{
    /** 
	\brief Target mass
      
	Use negative values to indicate a mass measured relative to the
	maximum mass. For example, if the EOS has a maximum mass of 2.0, then
	-0.15 will give the profile of a 1.85 solar mass star.
    */
    double tmass;
    //@}

    /// \name Results

    //@{
    /** \brief Return the results data table
     */
    table_units &get_results() {return out_table;}
    //@}
  
    //--------------------------------------------
    // Function prototypes

    tov_solve();

    virtual ~tov_solve();
    
    /// \name Actual solution of equations
    //@{
    /// Calculate the mass vs. radius curve
    int mvsr();
    /// Calculate the profile of a star with fixed mass
    int fixed(double d_tmass);
    /// Calculate the profile of the maximum mass star
    int max();
    //@}

    /** 
	\brief Check the solution (unfinished)
    */
    int solution_check();

    /// Set units
    int set_units(double s_efactor=1.0, double s_pfactor=1.0, 
		  double s_nbfactor=1.0);

    /** 
	\brief Set units
	
	Valid entries for the units of energy density and pressure are:
	- "g/cm^3"
	- "erg/cm^3"
	- "MeV/fm^3"
	- "fm^-4"
	- "Msun/km^3" (i.e. solar masses per cubic kilometer)
	
	Valid entries for the units of baryon density are:
	- "m^-3"
	- "cm^-3"
	- "fm^-3"
    */
    int set_units(std::string eunits="", std::string punits="", 
		  std::string nunits="");

    /** 
	\brief Set maximum storage for integration
	
	The variable \c s_kmax is the maximum number of radial
	integration stepsk while \c s_maxoutsize is the maximum number
	of points that will be output for any profile.
	
	If \c s_kmax is less than zero, there is no limit on
	the number of radial steps.
    */
    int set_kmax(int s_maxoutsize=400, int s_kmax=40000);

    /// Set the EOS to use
    int set_eos(tov_eos &ter) {
      te=&ter;
      eos_set=true;
      return 0;
    }

    /** 
	\brief Set solver
    */
    int set_mroot(mroot<int,mm_funct<int> > &s_mrp) 
    {mroot_ptr=&s_mrp; return 0;};
    
    /** 
	\brief Set minimizer
    */
    int set_minimize(minimize<int,funct<int> > &s_mp) 
    {min_ptr=&s_mp; return 0;};

    /** 
	\brief Set the adaptive stepper
    */
    int set_stepper(adapt_step<int,ode_funct<int> > &sap)
    {as_ptr=&sap; return 0;};

    /// The default minimizer
    gsl_min_brent<int,funct<int> > def_min;
    
    /// The default solver
    gsl_mroot_hybrids<int,mm_funct<int> > def_solver;

    /// The default adaptive stepper
    gsl_astep<int,ode_funct<int> > def_stepper;

    /// If true, compute the angular velocity (default false)
    bool compute_ang_vel;
    
    /// The angular velocity
    double cap_omega;

    /// The schwarzchild radius in km
    double schwarz_km;
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /** 
	\brief Ensure \c col does not match strings in \c cnames
    */
    int make_unique_name(std::string &col, 
			 std::vector<std::string> &cnames);
    
    /// The EOS
    tov_eos *te;

    /// True if the EOS has been set
    bool eos_set;

    /** \brief Define some necessary I/O objects

	\future Is this really required?
     */
    base_ioc bio;

    /// \name User EOS
    //@{
    /// Units for energy density
    std::string eunits;
    /// Units for pressure
    std::string punits;
    /// Units for baryon density
    std::string nunits;
    /// unit conversion factor for energy density (default 1.0)
    double efactor;
    /// unit conversion factor for pressure (default 1.0)
    double pfactor;
    /// unit conversion factor for baryon density (default 1.0)
    double nfactor;
    //@}
    
    /// maximum storage size (default 40000)
    int kmax;
    /// maximum size of output file (default 400)
    int maxoutsize;

    /** 
	\brief Smallest allowed pressure for integration (default: -100)
   
	This can't be much
	smaller since we need to compute numbers near exp(-presmin)
    */
    double presmin;

    /// \name Integration storage
    //@{
    ovector rky[6], rkx, rkdydx[6];
    //@}
    
    /// The output table
    table_units out_table;

    /// The solver
    mroot<int,mm_funct<int> > *mroot_ptr;

    /// The minimizer
    minimize<int,funct<int> > *min_ptr;
    
    /// The default adaptive stepper
    adapt_step<int,ode_funct<int> > *as_ptr;

    /// The ODE step function
    virtual int derivs(double x, size_t nv, const ovector_base &y,
		       ovector_base &dydx, int &pa);
    
    /// The ODE step function for the angular velocity
    virtual int derivs_ang_vel(double x, size_t nv, const ovector_base &y,
			       ovector_base &dydx, int &pa);
    
    /// Output a stellar profile
    virtual int profile_out(double xx);
    
    /// The minimizer function to compute the maximum mass
    virtual double maxfun(double maxx, int &pa);

    /** 
	\brief The solver function to compute the stellar profile

     */
    virtual int starfun(size_t ndvar, const ovector_base &ndx, 
			ovector_base &ndy, int &pa);
    
    /// Compute the angular velocity
    int ang_vel();

    /// Interpolation object for derivs_ang_vel()
    smart_interp<> smi;

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif


