/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_INTE_QAWC_H
#define O2SCL_GSL_INTE_QAWC_H

#include <o2scl/inte.h>
#include <o2scl/gsl_inte_qag_b.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** 
      \brief Chebyshev integration base class (GSL)

      This class provides the basic Chebyshev integration functions
      for use in the GSL-based integration classes which
      require them.
      
   */
  template<class param_t, class func_t> class gsl_inte_cheb : 
  public gsl_inte_transform<param_t,func_t> {
    
  public:
    
    /// Compute the Chebyshev moments
    void compute_moments(double cc, double *moment) {
      size_t k;
      
      double a0 = log (fabs ((1.0 - cc) / (1.0 + cc)));
      double a1 = 2 + a0 * cc;

      moment[0] = a0;
      moment[1] = a1;

      for (k = 2; k < 25; k++) {
	double a2;
	
	if ((k % 2) == 0) {
	  a2 = 2.0 * cc * a1 - a0;
	} else {
	  const double km1 = k - 1.0;
	  a2 = 2.0 * cc * a1 - a0 - 4.0 / (km1 * km1 - 1.0);
	}
	
	moment[k] = a2;
	
	a0 = a1;
	a1 = a2;
      }
    }

    /** \brief Perform the integration
	
	piessens,robert,appl. math. & progr. div. - k.u.leuven
	de doncker,elise,appl. math. & progr. div. - k.u.leuven
	
	this routine computes the chebyshev series expansion
	of degrees 12 and 24 of a function using a
	fast fourier transform method
	f(x) = sum(k=1,..,13) (cheb12(k)*t(k-1,x)),
	f(x) = sum(k=1,..,25) (cheb24(k)*t(k-1,x)),
	where t(k,x) is the chebyshev polynomial of degree k.


	x      - double precision
	vector of dimension 11 containing the
	values cos(k*pi/24), k = 1, ..., 11
	
	fval   - double precision
	vector of dimension 25 containing the
	function values at the points
	(b+a+(b-a)*cos(k*pi/24))/2, k = 0, ...,24,
	where (a,b) is the approximation interval.
	fval(1) and fval(25) are divided by two
	(these values are destroyed at output).
	
	on return
	cheb12 - double precision
	vector of dimension 13 containing the
	chebyshev coefficients for degree 12
	
	cheb24 - double precision
	vector of dimension 25 containing the
	chebyshev coefficients for degree 24

     */
    void gsl_integration_qcheb(func_t &f, double a, double b, 
			       double *cheb12, double *cheb24,
			       param_t &pa) {
      size_t i;
      double fval[25], v[12];
      
      /* These are the values of cos(pi*k/24) for k=1..11 needed for the
	 Chebyshev expansion of f(x) */
      
      const double x[11] = { 0.9914448613738104,     
			     0.9659258262890683,
			     0.9238795325112868,     
			     0.8660254037844386,
			     0.7933533402912352,     
			     0.7071067811865475,
			     0.6087614290087206,     
			     0.5000000000000000,
			     0.3826834323650898,     
			     0.2588190451025208,
			     0.1305261922200516 };
  
      const double center = 0.5 * (b + a);
      const double half_length =  0.5 * (b - a);

      double y1, y2, y3;
      f(b,y1,pa);
      f(center,y2,pa);
      f(a,y3,pa);
      fval[0] = 0.5 * y1;
      fval[12] = y2;
      fval[24] = 0.5 * y3;

      for (i = 1; i < 12; i++) {
	const size_t j = 24 - i;
	const double u = half_length * x[i-1];
	double yp, ym;
	f(center+u,yp,pa);
	f(center-u,ym,pa);
	fval[i] = yp;
	fval[j] = ym;
      }

      for (i = 0; i < 12; i++) {
	const size_t j = 24 - i;
	v[i] = fval[i] - fval[j];
	fval[i] = fval[i] + fval[j];
      }

      {
	const double alam1 = v[0] - v[8];
	const double alam2 = x[5] * (v[2] - v[6] - v[10]);

	cheb12[3] = alam1 + alam2;
	cheb12[9] = alam1 - alam2;
      }

      {
	const double alam1 = v[1] - v[7] - v[9];
	const double alam2 = v[3] - v[5] - v[11];
	{
	  const double alam = x[2] * alam1 + x[8] * alam2;

	  cheb24[3] = cheb12[3] + alam;
	  cheb24[21] = cheb12[3] - alam;
	}

	{
	  const double alam = x[8] * alam1 - x[2] * alam2;
	  cheb24[9] = cheb12[9] + alam;
	  cheb24[15] = cheb12[9] - alam;
	}
      }

      {
	const double part1 = x[3] * v[4];
	const double part2 = x[7] * v[8];
	const double part3 = x[5] * v[6];
    
	{
	  const double alam1 = v[0] + part1 + part2;
	  const double alam2 = x[1] * v[2] + part3 + x[9] * v[10];
      
	  cheb12[1] = alam1 + alam2;
	  cheb12[11] = alam1 - alam2;
	}
    
	{
	  const double alam1 = v[0] - part1 + part2;
	  const double alam2 = x[9] * v[2] - part3 + x[1] * v[10];
	  cheb12[5] = alam1 + alam2;
	  cheb12[7] = alam1 - alam2;
	}
      }

      {
	const double alam = (x[0] * v[1] + x[2] * v[3] + x[4] * v[5]
			     + x[6] * v[7] + x[8] * v[9] + x[10] * v[11]);
	cheb24[1] = cheb12[1] + alam;
	cheb24[23] = cheb12[1] - alam;
      }

      {
	const double alam = (x[10] * v[1] - x[8] * v[3] + x[6] * v[5] 
			     - x[4] * v[7] + x[2] * v[9] - x[0] * v[11]);
	cheb24[11] = cheb12[11] + alam;
	cheb24[13] = cheb12[11] - alam;
      }

      {
	const double alam = (x[4] * v[1] - x[8] * v[3] - x[0] * v[5] 
			     - x[10] * v[7] + x[2] * v[9] + x[6] * v[11]);
	cheb24[5] = cheb12[5] + alam;
	cheb24[19] = cheb12[5] - alam;
      }

      {
	const double alam = (x[6] * v[1] - x[2] * v[3] - x[10] * v[5] 
			     + x[0] * v[7] - x[8] * v[9] - x[4] * v[11]);
	cheb24[7] = cheb12[7] + alam;
	cheb24[17] = cheb12[7] - alam;
      }

      for (i = 0; i < 6; i++) {
	const size_t j = 12 - i;
	v[i] = fval[i] - fval[j];
	fval[i] = fval[i] + fval[j];
      }

      {
	const double alam1 = v[0] + x[7] * v[4];
	const double alam2 = x[3] * v[2];

	cheb12[2] = alam1 + alam2;
	cheb12[10] = alam1 - alam2;
      }

      cheb12[6] = v[0] - v[4];

      {
	const double alam = x[1] * v[1] + x[5] * v[3] + x[9] * v[5];
	cheb24[2] = cheb12[2] + alam;
	cheb24[22] = cheb12[2] - alam;
      }

      {
	const double alam = x[5] * (v[1] - v[3] - v[5]);
	cheb24[6] = cheb12[6] + alam;
	cheb24[18] = cheb12[6] - alam;
      }

      {
	const double alam = x[9] * v[1] - x[5] * v[3] + x[1] * v[5];
	cheb24[10] = cheb12[10] + alam;
	cheb24[14] = cheb12[10] - alam;
      }

      for (i = 0; i < 3; i++) {
	const size_t j = 6 - i;
	v[i] = fval[i] - fval[j];
	fval[i] = fval[i] + fval[j];
      }

      cheb12[4] = v[0] + x[7] * v[2];
      cheb12[8] = fval[0] - x[7] * fval[2];

      {
	const double alam = x[3] * v[1];
	cheb24[4] = cheb12[4] + alam;
	cheb24[20] = cheb12[4] - alam;
      }

      {
	const double alam = x[7] * fval[1] - fval[3];
	cheb24[8] = cheb12[8] + alam;
	cheb24[16] = cheb12[8] - alam;
      }

      cheb12[0] = fval[0] + fval[2];

      {
	const double alam = fval[1] + fval[3];
	cheb24[0] = cheb12[0] + alam;
	cheb24[24] = cheb12[0] - alam;
      }

      cheb12[12] = v[0] - v[2];
      cheb24[12] = cheb12[12];

      for (i = 1; i < 12; i++) {
	cheb12[i] *= 1.0 / 6.0;
      }

      cheb12[0] *= 1.0 / 12.0;
      cheb12[12] *= 1.0 / 12.0;

      for (i = 1; i < 24; i++) {
	cheb24[i] *= 1.0 / 12.0;
      }

      cheb24[0] *= 1.0 / 24.0;
      cheb24[24] *= 1.0 / 24.0;
    }

  };

  /** 
      \brief Adaptive Cauchy principal value integration (GSL)

      The location of the singularity must be specified before-hand in
      gsl_inte_qawc::s, and the singularity must not be at one of the
      endpoints. Note that when integrating a function of the form
      \f$ \frac{f(x)}{(x-s)} \f$, the denominator \f$ (x-s) \f$ must 
      not be specified in the argument \c func to integ(). This is
      different from how the \ref cern_cauchy operates.

      The number of subdivisions of the original interval which 
      this class is allowed to make is dictated by the workspace
      size for the integration class, which can be set using
      \ref gsl_inte_table::set_wkspace() . 
      
      \future Make cern_cauchy and this class consistent in the
      way which they require the user to provide the denominator
      in the integrand
  */
  template<class param_t, class func_t> class gsl_inte_qawc : 
  public gsl_inte_cheb<param_t,func_t> {
    
  public:

    gsl_inte_qawc() {
    }

    virtual ~gsl_inte_qawc() {}
      
    /** \brief Integrate function \c func from \c a to \c b.
     */
    virtual double integ(func_t &func, double a, double b, param_t &pa) {
      double res, err;
      integ_err(func,a,b,pa,res,err);
      this->interror=err;
      return res;
    }
    
    /// The singularity
    double s;

    /** \brief Integrate function \c func from \c a to \c b and place
	the result in \c res and the error in \c err
    */
    virtual int integ_err(func_t &func, double a, double b, 
			  param_t &pa, double &res, double &err2) {

      int status=qawc(func,a,b,s,this->tolx,this->tolf,this->wkspace,
		      &res,&err2,pa);
      
      return status;
      
    }

#ifndef DOXYGEN_INTERNAL

  protected:

    /** 
	\brief The full GSL integration routine called by integ_err()
    */
    int qawc(func_t &func, const double a, const double b, const double c,
	     const double epsabs, const double epsrel, const size_t limit,
	     double *result, double *abserr, param_t &pa) {
      
      double area, errsum;
      double result0, abserr0;
      double tolerance;
      size_t iteration = 0;
      int roundoff_type1 = 0, roundoff_type2 = 0, error_type = 0;
      int err_reliable;
      int sign = 1;
      double lower, higher;

      /* Initialize results */

      *result = 0;
      *abserr = 0;

      if (limit > this->w->limit) {
	std::string estr="Iteration limit exceeds workspace ";
	estr+="in gsl_inte_qawc::qawc().";
	O2SCL_ERR_RET(estr.c_str(),gsl_einval);
      }

      if (b < a)  {
	lower = b ;
	higher = a ;
	sign = -1 ;
      } else {
	lower = a;
	higher = b;
      }

      initialise (this->w, lower, higher);

      if (epsabs <= 0 && (epsrel < 50 * GSL_DBL_EPSILON || 
			  epsrel < 0.5e-28)) {
	std::string estr="Tolerance cannot be achieved with given ";
	estr+="value of 'tolx' and 'tolf' in gsl_inte_qawc::qawc().";
	O2SCL_ERR_RET(estr.c_str(),gsl_ebadtol);
      }

      if (c == a || c == b) {
	std::string estr="Cannot integrate with singularity on endpoint ";
	estr+="in gsl_inte_qawc::qawc().";
	O2SCL_ERR_RET(estr.c_str(),gsl_einval);
      }      

      /* perform the first integration */
      
      qc25c (func, lower, higher, c, &result0, &abserr0, &err_reliable, pa);

      set_initial_result (this->w, result0, abserr0);

      /* Test on accuracy, use 0.01 relative error as an extra safety
	 margin on the first iteration (ignored for subsequent iterations) 
      */
      tolerance = GSL_MAX_DBL (epsabs, epsrel * fabs (result0));

      if (abserr0 < tolerance && abserr0 < 0.01 * fabs(result0)) {
	*result = sign * result0;
	*abserr = abserr0;

	return GSL_SUCCESS;

      } else if (limit == 1) {

	*result = sign * result0;
	*abserr = abserr0;

	std::string estr="A maximum of 1 iteration was insufficient ";
	estr+="in gsl_inte_qawc::qawc().";
	O2SCL_ERR_RET(estr.c_str(),gsl_emaxiter);
      }

      area = result0;
      errsum = abserr0;

      iteration = 1;

      do {

	double a1, b1, a2, b2;
	double a_i, b_i, r_i, e_i;
	double area1 = 0, area2 = 0, area12 = 0;
	double error1 = 0, error2 = 0, error12 = 0;
	int err_reliable1, err_reliable2;

	/* Bisect the subinterval with the largest error estimate */

	retrieve (this->w, &a_i, &b_i, &r_i, &e_i);

	a1 = a_i; 
	b1 = 0.5 * (a_i + b_i);
	a2 = b1;
	b2 = b_i;

	if (c > a1 && c <= b1) {
	  b1 = 0.5 * (c + b2) ;
	  a2 = b1;
	} else if (c > b1 && c < b2) {
	  b1 = 0.5 * (a1 + c) ;
	  a2 = b1;
	}

	qc25c (func, a1, b1, c, &area1, &error1, &err_reliable1, pa);
	qc25c (func, a2, b2, c, &area2, &error2, &err_reliable2, pa);

	area12 = area1 + area2;
	error12 = error1 + error2;

	errsum += (error12 - e_i);
	area += area12 - r_i;

	if (err_reliable1 && err_reliable2) {
	  double delta = r_i - area12;

	  if (fabs (delta) <= 1.0e-5 * fabs (area12) && 
	      error12 >= 0.99 * e_i) {
	    roundoff_type1++;
	  }
	  if (iteration >= 10 && error12 > e_i) {
	    roundoff_type2++;
	  }
	}

	tolerance = GSL_MAX_DBL (epsabs, epsrel * fabs (area));

	if (errsum > tolerance) {
	  if (roundoff_type1 >= 6 || roundoff_type2 >= 20) {
	    error_type = 2;   /* round off error */
	  }

	  /* set error flag in the case of bad integrand behaviour at
	     a point of the integration range */

	  if (this->subinterval_too_small (a1, a2, b2)) {
	    error_type = 3;
	  }
	}

	update (this->w, a1, b1, area1, error1, a2, b2, area2, error2);

	retrieve (this->w, &a_i, &b_i, &r_i, &e_i);

	iteration++;

      } while (iteration < limit && !error_type && errsum > tolerance);

      *result = sign * sum_results (this->w);
      *abserr = errsum;

      if (errsum <= tolerance) {
	return GSL_SUCCESS;
      } else if (error_type == 2) {
	std::string estr="Roundoff error prevents tolerance ";
	estr+="from being achieved in gsl_inte_qawc::qawc().";
	O2SCL_ERR_RET(estr.c_str(),gsl_eround);
      } else if (error_type == 3) {
	std::string estr="Bad integrand behavior ";
	estr+=" in gsl_inte_qawc::qawc().";
	O2SCL_ERR_RET(estr.c_str(),gsl_esing);
      } else if (iteration == limit) {
	std::string estr="Maximum number of subdivisions reached";
	estr+=" in gsl_inte_qawc::qawc().";
	O2SCL_ERR_RET(estr.c_str(),gsl_emaxiter);
      } else {
	std::string estr="Could not integrate function in gsl_inte_qawc::";
	estr+="qawc() (it may have returned a non-finite result).";
	O2SCL_ERR_RET(estr.c_str(),gsl_efailed);
      }

      // No return statement needed since the above if statement
      // always forces a return
    }

    /// 25-point quadrature for Cauchy principal values
    void qc25c(func_t &func, double a, double b, double c, 
	       double *result, double *abserr, int *err_reliable,
	       param_t &pa) {
      double cc = (2 * c - b - a) / (b - a);
      
      if (fabs (cc) > 1.1) {
	double resabs, resasc;
	    
	double fv1[8], fv2[8];
	gsl_integration_qk_o2scl(func,8,o2scl_inte_qag_coeffs::qk15_xgk,
				 o2scl_inte_qag_coeffs::qk15_wg,
				 o2scl_inte_qag_coeffs::qk15_wgk,
				 fv1,fv2,a,b,result,abserr,&resabs,&resasc,
				 pa);
      
	if (*abserr == resasc) {
	  *err_reliable = 0;
	} else {
	  *err_reliable = 1;
	}

	return;
      } else {

	double cheb12[13], cheb24[25], moment[25];
	double res12 = 0, res24 = 0;
	size_t i;
	this->gsl_integration_qcheb (func, a, b, cheb12, cheb24, pa);
	this->compute_moments (cc, moment);
	  
	for (i = 0; i < 13; i++) {
	  res12 += cheb12[i] * moment[i];
	}
	  
	for (i = 0; i < 25; i++) {
	  res24 += cheb24[i] * moment[i];
	}
	  
	*result = res24;
	*abserr = fabs(res24 - res12) ;
	*err_reliable = 0;

	return;
      }
    }

    /// Add the singularity to the function
    virtual double transform(func_t &func, double x, param_t &pa) {
      double y;
      func(x,y,pa);
      return y/(x-s);
    }

#endif
  
    /// Return string denoting type ("gsl_inte_qawc")
    const char *type() { return "gsl_inte_qawc"; }
  
  };

#ifndef DOXYGENP
}
#endif

#endif
