/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#include <o2scl/multi_funct.h>
#include <o2scl/gsl_mmin_conp.h>
#include <o2scl/test_mgr.h>

using namespace std;
using namespace o2scl;

// Test functions for ovectors
double minfun(size_t n, const ovector_base &x, int &pa) {
  double ret;
  ret=x[0]*x[0]+(x[1]-2.0)*(x[1]-2.0)+3.0;
  return(ret);
}

int minfund(size_t n, ovector_base &x, ovector_base &g, int &pa) {
  double ret;
  g[0]=2.0*x[0];
  g[1]=2.0*(x[1]-2.0);
  return 0;
}

// Test functions for arrays
int minfun2(size_t n, const double x[2], double &y, int &pa) {
  y=x[0]*x[0]+(x[1]-2.0)*(x[1]-2.0)+3.0;
  return 0;
}

int minfund2(size_t n, double x[2], double g[2], int &pa) {
  double ret;
  g[0]=2.0*x[0];
  g[1]=2.0*(x[1]-2.0);
  return 0;
}

// Test functions in GSL form
double my_f(const gsl_vector *v, void *pa) {
  double x = gsl_vector_get(v,0);
  double y = gsl_vector_get(v,1);
  return x*x+(y-2.0)*(y-2.0)+3.0;
}

void my_df(const gsl_vector *v, void *params, gsl_vector *df) {
  double x = gsl_vector_get(v,0);
  double y = gsl_vector_get(v,1);
  
  gsl_vector_set(df,0,2.0*x);
  gsl_vector_set(df,1,2.0*(y-2.0));

}

void my_fdf(const gsl_vector *x, void *params, double *f, gsl_vector *df) {
  *f = my_f(x,params);
  my_df(x,params,df);
}

// To demonstrate that the user can specify their own automatic
// gradient class
template<class param_t, class func_t, class vec_t> class simple_grad_new :
  public simple_grad<param_t,func_t,vec_t> {
public:
  virtual ~simple_grad_new() {}
  virtual int operator()(size_t nv, vec_t &x, vec_t &g, param_t &pa) {
    std::cout << '.' << std::flush;
    return simple_grad<param_t,func_t,vec_t>::operator()(nv,x,g,pa);
  }
};

// To demonstrate that the user can specify their own automatic
// gradient class for arrays
template<class param_t, class func_t, size_t nv> class simple_grad_array_new :
  public simple_grad_array<param_t,func_t,nv> {
public:
  virtual ~simple_grad_array_new() {}
  virtual int operator()(size_t nvar, double x[nv], double g[nv], 
			 param_t &pa) {
    std::cout << '.' << std::flush;
    return simple_grad_array<param_t,func_t,nv>::operator()(nv,x,g,pa);
  }
};

int main(void) {
  test_mgr t;
  t.set_output_level(1);
  def_err_hnd.set_mode(2);

  double min;
  ovector x(2);
  double xa[2];
  int vp=0;
  int ret;
  gsl_mmin_conp<int> g;
  gsl_mmin_conp_array<int,2> g2;

  cout.setf(ios::scientific);

  multi_funct_fptr_noerr<int> mf(minfun);
  grad_funct_fptr<int> mfd(minfund);
  multi_vfunct_fptr<int,2> mf2(minfun2);
  grad_vfunct_fptr<int,2> mfd2(minfund2);

  /*
    Note that the ovector and array versions exactly match on my
    desktop, but don't quite match on my laptop. This distinction is
    demonstrated by the "conp_compare" target in the Makefile.

    On the desktop, the ovector, array, and GSL versions match 
    perfectly.
  */

  // Normal version without gradient
  
  x[0]=1.0;
  x[1]=1.0;
  ret=g.mmin(2,x,min,vp,mf);
  cout << g.last_ntrial << endl;
  cout << ret << " " << min << " " << x[0] << " " << x[1] << endl;
  t.test_gen(ret==0,"gsl_mmin_conp 0");
  t.test_abs(x[0],0.0,1.0e-4,"gsl_mmin_conp 1");
  t.test_rel(x[1],2.0,2.0e-4,"gsl_mmin_conp 2");
  cout << endl;

  // Array version without gradient

  xa[0]=1.0;
  xa[1]=1.0;
  ret=g2.mmin(2,xa,min,vp,mf2);
  cout << g2.last_ntrial << endl;
  cout << ret << " " << min << " " << xa[0] << " " << xa[1] << endl;
  t.test_gen(ret==0,"gsl_mmin_conp array 0");
  t.test_abs(xa[0],0.0,1.0e-4,"gsl_mmin_conp array 1");
  t.test_rel(xa[1],2.0,2.0e-4,"gsl_mmin_conp array 2");
  t.test_abs(x[0],xa[0],1.0e-9,"gsl_mmin_conp compare 1");
  t.test_rel(x[1],xa[1],1.0e-6,"gsl_mmin_conp compare 2");
  cout << endl;

  // Show that we can use a user-specified automatic gradient object
  gsl_mmin_conp<int,multi_funct<int>,ovector_base,ovector,
    ovector_alloc,grad_funct<int,ovector_base>,
    gradient<int,multi_funct<int>,ovector_base>,
    simple_grad_new<int,multi_funct<int>,ovector_base> > g3;

  // Emacs has trouble with tabification
#ifdef O2SCL_NEVER_DEFINED
} {
#endif
  
  gsl_mmin_conp<int,multi_vfunct<int,2>,
    double[2],double[2],array_alloc<double[2]>,grad_vfunct<int,2>,
    gradient_array<int,multi_vfunct<int,2>,2>,
    simple_grad_array_new<int,multi_vfunct<int,2>,2> > g4;

  // Emacs has trouble with tabification
#ifdef O2SCL_NEVER_DEFINED
} {
#endif

  // Normal version with user auto gradient

  x[0]=1.0;
  x[1]=1.0;
  ret=g3.mmin(2,x,min,vp,mf);
  cout << g3.last_ntrial << endl;
  cout << ret << " " << min << " " << x[0] << " " << x[1] << endl;
  t.test_gen(ret==0,"gsl_mmin_conp grad 0");
  t.test_abs(x[0],0.0,1.0e-4,"gsl_mmin_conp grad 1");
  t.test_rel(x[1],2.0,2.0e-4,"gsl_mmin_conp grad 2");
  cout << endl;

  // Array version with user auto gradient

  xa[0]=1.0;
  xa[1]=1.0;
  ret=g4.mmin(2,xa,min,vp,mf2);
  cout << g4.last_ntrial << endl;
  cout << ret << " " << min << " " << xa[0] << " " << xa[1] << endl;
  t.test_gen(ret==0,"gsl_mmin_conp arr grad 0");
  t.test_abs(xa[0],0.0,1.0e-4,"gsl_mmin_conp arr grad 1");
  t.test_rel(xa[1],2.0,2.0e-4,"gsl_mmin_conp arr grad 2");
  t.test_abs(x[0],xa[0],1.0e-9,"gsl_mmin_conp compare 3");
  t.test_rel(x[1],xa[1],1.0e-6,"gsl_mmin_conp compare 4");
  cout << endl;

  // Normal version with gradient

  x[0]=1.0;
  x[1]=1.0;
  ret=g.mmin_de(2,x,min,vp,mf,mfd);
  cout << g.last_ntrial << endl;
  cout << ret << " " << min << " " << x[0] << " " << x[1] << endl;
  t.test_gen(ret==0,"gsl_mmin_conp grad 0");
  t.test_abs(x[0],0.0,1.0e-4,"gsl_mmin_conp grad 1");
  t.test_rel(x[1],2.0,2.0e-4,"gsl_mmin_conp grad 2");
  cout << endl;

  // Array version with gradient

  xa[0]=1.0;
  xa[1]=1.0;
  ret=g2.mmin_de(2,xa,min,vp,mf2,mfd2);
  cout << g2.last_ntrial << endl;
  cout << ret << " " << min << " " << xa[0] << " " << xa[1] << endl;
  t.test_gen(ret==0,"gsl_mmin_conp arr grad 0");
  t.test_abs(xa[0],0.0,1.0e-4,"gsl_mmin_conp arr grad 1");
  t.test_rel(xa[1],2.0,2.0e-4,"gsl_mmin_conp arr grad 2");
  t.test_abs(x[0],xa[0],1.0e-15,"gsl_mmin_conp compare 3");
  t.test_rel(x[1],xa[1],1.0e-12,"gsl_mmin_conp compare 4");
  cout << endl;
  
  // GSL version with gradient
  {
    int iter = 0;
    int status;
    
    const gsl_multimin_fdfminimizer_type *T;
    gsl_multimin_fdfminimizer *s;
    
    gsl_vector *gx;
    gsl_multimin_function_fdf my_func;
    
    my_func.f = &my_f;
    my_func.df = &my_df;
    my_func.fdf = &my_fdf;
    my_func.n = 2;
    my_func.params = 0;
    
    gx = gsl_vector_alloc (2);
    gsl_vector_set(gx,0,1.0);
    gsl_vector_set(gx,1,1.0);
    T = gsl_multimin_fdfminimizer_conjugate_pr;
    s = gsl_multimin_fdfminimizer_alloc (T, 2);
    
    gsl_multimin_fdfminimizer_set(s,&my_func,gx,0.01,1e-4);
    
    do {
      iter++;
      status = gsl_multimin_fdfminimizer_iterate(s);
      /*
	cout << "Iter: " << iter << " " << gsl_vector_get(s->x,0) << " "
	<< gsl_vector_get(s->x,1) << " " 
	<< gsl_vector_get(s->gradient,0) << " "
	<< gsl_vector_get(s->gradient,1) << " "
	<< endl;
      */
	
      if (status)
	break;
      
      status = gsl_multimin_test_gradient(s->gradient,1e-4);
      
    } while (status == GSL_CONTINUE && iter < 100);
    
    cout << iter << endl;
    cout << status << " " << s->f << " "
	 << gsl_vector_get(s->x,0) << " "
	 << gsl_vector_get(s->x,1) << endl;
    t.test_rel(s->f,min,1.0e-12,"conp compare 5");
    t.test_abs(gsl_vector_get(s->x,0),xa[0],1.0e-15,"conp compare 6");
    t.test_rel(gsl_vector_get(s->x,1),xa[1],1.0e-12,"conp compare 7");
    t.test_gen(iter==g.last_ntrial,"conp compare 8");
    t.test_gen(iter==g2.last_ntrial,"conp compare 9");
    
    gsl_multimin_fdfminimizer_free(s);
    gsl_vector_free(gx);
  }

  t.report();
  return 0;
}
 
