/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

#ifndef O2SCL_NONADAPT_STEP_H
#define O2SCL_NONADAPT_STEP_H

#include <o2scl/adapt_step.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_odeiv.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief An non-adaptive stepper implementation of \ref adapt_step.

      This class simply calls the specified ODE stepper without any
      attempt to modify the size of the step, and is primarily useful
      to allow for simple comparisons between adaptive and
      non-adaptive solution. To modify the ODE stepper which is used,
      use the adapt_step::set_step().
      
      \future Modify so that memory allocation/deallocation is only
      performed when necessary
  */
  template<class param_t, class func_t=ode_funct<param_t>, 
    class vec_t=ovector_base, class alloc_vec_t=ovector, 
    class alloc_t=ovector_alloc> class nonadapt_step : 
  public adapt_step<param_t,func_t,vec_t,alloc_vec_t,alloc_t> {

#ifndef DOXYGEN_INTERNAL

  protected:

  /// The allocated vector size
  size_t msize;

  /** 
      \brief Internal storage for dydx
      
      \comment
      This isn't named dydx so that we can name function
      arguments dydx in astep_derivs().
      \endcomment
  */
  alloc_vec_t dydx_int;

  /// Memory allocator for objects of type \c alloc_vec_t
  alloc_t ao;
      
#endif
    
    public:
    
    nonadapt_step()  {
      msize=0;
    }
      
    virtual ~nonadapt_step() {
      if (msize>0) {
	ao.free(dydx_int);
      }
    }

    /** 
	\brief Make an adaptive integration step of the system 
	\c derivs 
	  
	This attempts to take a step of size \c h from the point \c
	x of an \c n-dimensional system \c derivs starting with \c
	y. On exit, \c x and \c y contain the new values at the end
	of the step, \c h contains the size of the step, \c dydx_out
	contains the derivative at the end of the step, and \c yerr
	contains the estimated error at the end of the step.

	If the base stepper returns a non-zero value, that
	value is passed on as the return value of <tt>astep()</tt>,
	though the input \c y may still have been modified by
	the base stepper.

    */
    virtual int astep(double &x, double &h, double xmax,
		      size_t n, vec_t &y, vec_t &u_dydx_out,
		      vec_t &yerr, param_t &pa, 
		      func_t &derivs) {

      if (n!=msize) {
	ao.allocate(dydx_int,n);
      }
      
      derivs(x,n,y,dydx_int,pa);
      int ret=this->stepp->step(x,h,n,y,dydx_int,y,yerr,
				u_dydx_out,pa,derivs);
      x+=h;

      return ret;
    }
      
    /** 
	\brief Make an adaptive integration step of the system 
	\c derivs with derivatives
	  
	This attempts to take a step of size \c h from the point \c
	x of an \c n-dimensional system \c derivs starting with \c y
	and given the initial derivatives \c dydx. On exit, \c x, \c
	y and \c dydx contain the new values at the end of the step,
	\c h contains the size of the step, \c dydx
	contains the derivative at the end of the step, and \c yerr
	contains the estimated error at the end of the step.

	If the base stepper returns a non-zero value, that value is
	passed on as the return value of <tt>astep()</tt>, though the
	inputs \c y and \c dydx may still have been modified by the
	base stepper.

    */
    virtual int astep_derivs(double &x, double &h, double xmax,
			     size_t n, vec_t &y, vec_t &dydx, 
			     vec_t &yerr, param_t &pa, 
			     func_t &derivs) {
      
      int ret=this->stepp->step(x,h,n,y,dydx,y,yerr,dydx,pa,derivs);
      x+=h;

      return ret;
    }
  
  };
  
#ifndef DOXYGENP
}
#endif

#endif
