/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ODE_IT_SOLVE_H
#define O2SCL_ODE_IT_SOLVE_H

#include <iostream>

#include <gsl/gsl_linalg.h>
#include <gsl/gsl_ieee_utils.h>

#include <o2scl/misc.h>
#include <o2scl/array.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/omatrix_tlate.h>
#include <o2scl/uvector_tlate.h>
#include <o2scl/umatrix_tlate.h>
#include <o2scl/test_mgr.h>
#include <o2scl/linear_solver.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /// Function class for ode_it_solve
  template<class vec_t=o2scl::ovector_base> class ode_it_funct {
  
  public:
  
  ode_it_funct() {}
  
  virtual ~ode_it_funct() {}
  
  /// Using \c x and \c y, return the value of function number \c ieq 
  virtual double operator()(size_t ieq, double x, vec_t &y) {
    return 0.0;
  }

  };

  /// Function pointer for ode_it_solve
  template<class vec_t=o2scl::ovector_base> 
    class ode_it_funct_fptr : public ode_it_funct<vec_t> {

  public:

  virtual ~ode_it_funct_fptr() {}

  /// Create using a function pointer
  ode_it_funct_fptr(double (*fp)(size_t, double, vec_t &)) {
    fptr=fp;
  }

  /// Using \c x and \c y, return the value of function number \c ieq 
  virtual double operator()(size_t ieq, double x, vec_t &y) {
    return fptr(ieq,x,y);
  }

  protected:

  ode_it_funct_fptr() {}

  /// The function pointer
  double (*fptr)(size_t ieq, double x, vec_t &y);
  };

  /// Member function pointer for ode_it_solve
  template<class tclass, class vec_t=o2scl::ovector_base>
    class ode_it_funct_mfptr : public ode_it_funct<vec_t> {
  public:

  virtual ~ode_it_funct_mfptr() {}

  /// Create using a class instance and member function
  ode_it_funct_mfptr(tclass *tp, 
		     double (tclass::*fp)(size_t, double, vec_t &)) {
    tptr=tp;
    fptr=fp;
  }

  /// Using \c x and \c y, return the value of function number \c ieq 
  virtual double operator()(size_t ieq, double x, vec_t &y) {
    return (*tptr.*fptr)(ieq,x,y);
  }

  protected:

  ode_it_funct_mfptr() {}

  /// The class pointer
  tclass *tptr;

  /// The member function pointer
  double (tclass::*fptr)(size_t ieq, double x, vec_t &y);

  };

  /** \brief ODE solver using a generic linear solver to solve 
      finite-difference equations

      \todo Implement as a child of ode_bv_solve ?
      \todo Max and average tolerance?
      \todo partial correction option?
  */
  template <class func_t, class vec_t, class mat_t, class matrix_row_t,
    class solver_vec_t, class solver_mat_t> class ode_it_solve {
  
  public:
  
    ode_it_solve() {
      h=1.0e-4;
      niter=30;
      tolf=1.0e-8;
      verbose=0;
    }

    virtual ~ode_it_solve() {}

    /// Set level of output (default 0)
    int verbose;

    /// Stepsize for finite differencing (default \f$ 10^{-4} \f$)
    double h;

    /// Tolerance (default \f$ 10^{-8} \f$)
    double tolf;
  
    /// Maximum number of iterations (default 30)
    size_t niter;
  
    /// Set the linear solver
    int set_solver(o2scl_linalg::linear_solver<solver_vec_t,solver_mat_t> &ls) {
      solver=&ls;
      return 0;
    }
    
    /// Solve \c derivs with boundary conditions \c left and \c right
    int solve(size_t ngrid, size_t neq, size_t nbleft, vec_t &x, mat_t &y,
	      func_t &derivs, func_t &left, func_t &right,
	      solver_mat_t &mat, solver_vec_t &rhs, solver_vec_t &dy) {

      // Store the functions for simple derivatives
      fd=&derivs;
      fl=&left;
      fr=&right;
    
      // Variable index
      size_t ix;

      // Number of RHS boundary conditions
      size_t nbright=neq-nbleft;

      // Number of variables
      size_t nvars=ngrid*neq;
    
      bool done=false;
      for(size_t it=0;done==false && it<niter;it++) {
      
	ix=0;
      
	mat.set_all(0.0);

	// Construct the entries corresponding to the LHS boundary
	for(size_t i=0;i<nbleft;i++) {
	  matrix_row_t yk(y,0);
	  rhs[ix]=-left(i,x[0],yk);
	  for(size_t j=0;j<neq;j++) {
	    int rxa=mat.set(ix,j,fd_left(i,j,x[0],yk));
	  }
	  ix++;
	}

	// Construct the matrix entries for the internal points
	for(size_t k=0;k<ngrid-1;k++) {
	  size_t kp1=k+1;
	  double tx=(x[kp1]+x[k])/2.0;
	  double dx=x[kp1]-x[k];
	  matrix_row_t yk(y,k);
	  matrix_row_t ykp1(y,k+1);
	
	  for(size_t i=0;i<neq;i++) {
	  
	    rhs[ix]=y[k][i]-y[kp1][i]+(x[kp1]-x[k])*
	      (derivs(i,tx,ykp1)+derivs(i,tx,yk))/2.0;
	  
	    size_t lhs=k*neq;
	    for(size_t j=0;j<neq;j++) {
	      int rxb=mat.set(ix,lhs+j,-fd_derivs(i,j,tx,yk)*dx/2.0);
	      int rxc=mat.set(ix,lhs+j+neq,-fd_derivs(i,j,tx,ykp1)*dx/2.0);
	      if (i==j) {
		int rxd=mat.set(ix,lhs+j,mat(ix,lhs+j)-1.0);
		int rxe=mat.set(ix,lhs+j+neq,mat(ix,lhs+j+neq)+1.0);
	      }
	    }

	    ix++;

	  }
	}
      
	// Construct the entries corresponding to the RHS boundary
	for(size_t i=0;i<nbright;i++) {
	  matrix_row_t ylast(y,ngrid-1);
	  size_t lhs=neq*(ngrid-1);
	
	  rhs[ix]=-right(i,x[ngrid-1],ylast);
	
	  for(size_t j=0;j<neq;j++) {
	    int rxf=mat.set(ix,lhs+j,fd_right(i,j,x[ngrid-1],ylast));
	  }
	  
	  ix++;

	}
      
	// Compute correction by calling the linear solver

	if (verbose>3) {
	  for(size_t i=0;i<nvars;i++) {
	    for(size_t j=0;j<nvars;j++) {
	      std::cout << mat[i][j] << " ";
	    }
	    std::cout << std::endl;
	  }
	  for(size_t i=0;i<nvars;i++) {
	    std::cout << rhs[i] << std::endl;
	  }
	}

	int ret=solver->solve(ix,mat,rhs,dy);

	if (verbose>3) {
	  for(size_t i=0;i<nvars;i++) {
	    std::cout << dy[i] << std::endl;
	  }
	  char ch;
	  std::cin >> ch;
	}
      
	// Apply correction and compute its size

	double res=0.0;
	ix=0;

	for(size_t igrid=0;igrid<ngrid;igrid++) {
	  for(size_t ieq=0;ieq<neq;ieq++) {
	    y[igrid][ieq]+=dy[ix];
	    res+=dy[ix]*dy[ix];
	    ix++;
	  }
	}

	if (verbose>0) {
	  // Since we're in the o2scl namespace, we explicitly
	  // specify std::sqrt() here
	  std::cout << it << " " << std::sqrt(res) << " " 
		    << tolf << std::endl;
	}
      
	// If the correction has become small enough, we're done
	if (std::sqrt(res)<=tolf) done=true;
      }

      if (done==false) {
	O2SCL_ERR_RET("Exceeded number of iterations in solve().",
		      o2scl::gsl_emaxiter);
      }

      return 0;
    }

  protected:
  
    /// \name Storage for functions
    //@{
    ode_it_funct<vec_t> *fl, *fr, *fd;
    //@}

    /// Solver
    o2scl_linalg::linear_solver<solver_vec_t,solver_mat_t> *solver;
  
    /// Compute the derivatives of the LHS boundary conditions
    double fd_left(size_t ieq, size_t ivar, double x, vec_t &y) {
      double ret, dydx;
    
      y[ivar]+=h;
      ret=(*fl)(ieq,x,y);
    
      y[ivar]-=h;
      ret-=(*fl)(ieq,x,y);
    
      ret/=h;
      return ret;
    }
  
    /// Compute the derivatives of the RHS boundary conditions
    double fd_right(size_t ieq, size_t ivar, double x, vec_t &y) {
      double ret, dydx;
    
      y[ivar]+=h;
      ret=(*fr)(ieq,x,y);
    
      y[ivar]-=h;
      ret-=(*fr)(ieq,x,y);
    
      ret/=h;
      return ret;
    }
  
    /// Compute the finite-differenced part of the differential equations
    double fd_derivs(size_t ieq, size_t ivar, double x, vec_t &y) {
      double ret, dydx;
    
      y[ivar]+=h;
      ret=(*fd)(ieq,x,y);
    
      y[ivar]-=h;
      ret-=(*fd)(ieq,x,y);
    
      ret/=h;
    
      return ret;
    }
  
  };

#ifndef DOXYGENP
}
#endif

#endif
