/*
  -------------------------------------------------------------------
  
  Copyright (C) 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_CONVERT_UNITS_H
#define O2SCL_CONVERT_UNITS_H

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <map>

#include <o2scl/err_hnd.h>
#include <o2scl/file_format.h>
#include <o2scl/misc.h>
#include <o2scl/constants.h>
#include <o2scl/string_conv.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Convert units

      Allow the user to convert between two different units after
      specifying a conversion factor. This class will also
      automatically combine two conversion factors to create a new
      unit conversion.

      Conversions are performed by the \ref convert() function and
      the conversion factors must be specified beforehand using the 
      \ref insert_cache() function.

      Example:
      \code
      convert_units cu;
      cu.insert_cache("in","cm",2.54);
      cout << "12 in is " << cu.convert("in","cm",12.0) << " cm. " << endl;
      \endcode
      
      \future Ideally, a real C++ API for the GNU units command
      would probably be better.
  */
  class convert_units {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// The type for caching unit conversions
    typedef struct {
      /// The input unit
      std::string f;
      /// The output unit
      std::string t;
      /// The conversion factor
      double c;
    } unit_t;

    /// The cache where unit conversions are stored
    std::map<std::string,unit_t,string_comp> mcache;
    
    /// The iterator type
    typedef std::map<std::string,unit_t,string_comp>::iterator miter;
      
#endif

  public:

    /// Verbosity (default 0)
    int verbose;

    /// (default false)
    bool use_gnu_units;

    /// (default true)
    bool err_on_fail;

    /// Default 'units'
    std::string units_cmd_string;

    convert_units() {
      verbose=0;
      use_gnu_units=false;
      units_cmd_string="units";
      err_on_fail=true;
    }

    virtual ~convert_units() {}

    /** \brief Return the value \c val after converting using units \c
	from and \c to
    */
    virtual double convert(std::string from, std::string to, double val);

    /// Manually insert a unit conversion into the cache
    int insert_cache(std::string from, std::string to, double conv);

    /// Manually remove a unit conversion into the cache
    int remove_cache(std::string from, std::string to);
    
    /// Print the present unit cache to std::cout
    int print_cache();

    /// Add conversion factors for energy equivalents
    int energy_conv();

  };

#ifndef DOXYGENP
}
#endif

#endif
