/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/twod_intp.h>

using namespace std;
using namespace o2scl;

twod_intp::twod_intp() {
  it_set=false;
  data_set=false;
  bimp1=&dim1;
  bimp2=&dim2;
}

twod_intp::~twod_intp() {
  free_data();
}

int twod_intp::free_data() {
  if (data_set) {
    size_t i;
    if (xfirst==true) {
      for(i=0;i<ny;i++)	{
	delete si[i];
	delete ari[i];
      }
      delete[] si;
      delete[] ari;
    } else {
      for(i=0;i<nx;i++)	{
	delete si[i];
	delete aci[i];
      }
      delete[] si;
      delete[] aci;
    }
    data_set=false;
  }
  return 0;
}

int twod_intp::set_data(size_t n_x, size_t n_y, ovector &x_grid,
			ovector &y_grid, omatrix &data, 
			bool x_first) {
  size_t i, j;
  
  // Free interpolation objects from previous data
  free_data();
  
  // Set new data
  xfirst=x_first;
  nx=n_x;
  ny=n_y;
  xfun=&x_grid;
  yfun=&y_grid;
  datap=&data;

  // Allocate new interpolation objects
  if (x_first==true) {
    si=new sm_interp_vec *[ny];
    ari=new omatrix_row *[ny];
    for(i=0;i<ny;i++) {
      ari[i]=new omatrix_row(*datap,i);
      si[i]=new sm_interp_vec(*bimp1,*bimp2,nx,*xfun,*(ari[i]));
    }
  } else {
    si=new sm_interp_vec *[nx];
    aci=new omatrix_col *[nx];
    for(i=0;i<nx;i++) {
      aci[i]=new omatrix_col(*datap,i);
      si[i]=new sm_interp_vec(*bimp1,*bimp2,ny,*yfun,*aci[i]);
    }
  }
  data_set=true;

  return 0;
}

int twod_intp::reset_interp() {

  if (data_set) {
    free_data();

    // Allocate new interpolation objects
    if (xfirst==true) {
      si=new sm_interp_vec *[ny];
      ari=new omatrix_row *[ny];
      for(size_t i=0;i<ny;i++) {
	ari[i]=new omatrix_row(*datap,i);
	si[i]=new sm_interp_vec(*bimp1,*bimp2,nx,*xfun,*ari[i]);
      }
    } else {
      si=new sm_interp_vec *[nx];
      aci=new omatrix_col *[nx];
      for(size_t i=0;i<nx;i++) {
	aci[i]=new omatrix_col(*datap,i);
	si[i]=new sm_interp_vec(*bimp1,*bimp2,ny,*yfun,*aci[i]);
      }
    }
    
    data_set=false;
  } else {
    O2SCL_ERR("Data not set in twod_intp::reset_interp().",gsl_einval);
    return gsl_einval;
  }
  return 0;
}

double twod_intp::interp(double x, double y) {
  if (data_set==false) {
    O2SCL_ERR("Data not set in twod_intp::interp().",gsl_efailed);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(size_t i=0;i<ny;i++) {
      icol[i]=si[i]->interp(x);
    }
    if (it_set) {
      sm_interp_vec siy(*bimp1,*bimp2,ny,*yfun,icol);
      result=siy.interp(y);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.interp(y);
    }
  } else {
    ovector icol(nx);
    for(size_t i=0;i<nx;i++) {
      icol[i]=si[i]->interp(y);
    }
    if (it_set) {
      sm_interp_vec six(*bimp1,*bimp2,nx,*xfun,icol);
      result=six.interp(x);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.interp(x);
    }
  }
  return result;
}

double twod_intp::deriv_x(double x, double y) {
  if (data_set==false) {
    O2SCL_ERR("Data not set in twod_intp::interp().",gsl_efailed);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(size_t i=0;i<ny;i++) {
      icol[i]=si[i]->deriv(x);
    }
    if (it_set) {
      sm_interp_vec siy(*bimp1,*bimp2,ny,*yfun,icol);
      result=siy.interp(y);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.interp(y);
    }
  } else {
    ovector icol(nx);
    for(size_t i=0;i<nx;i++) {
      icol[i]=si[i]->interp(y);
    }
    if (it_set) {
      sm_interp_vec six(*bimp1,*bimp2,nx,*xfun,icol);
      result=six.deriv(x);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.deriv(x);
    }
  }
  return result;
}

double twod_intp::deriv_xy(double x, double y) {
  if (data_set==false) {
    O2SCL_ERR("Data not set in twod_intp::interp().",gsl_efailed);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(size_t i=0;i<ny;i++) {
      icol[i]=si[i]->deriv(x);
    }
    if (it_set) {
      sm_interp_vec siy(*bimp1,*bimp2,ny,*yfun,icol);
      result=siy.deriv(y);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.deriv(y);
    }
  } else {
    ovector icol(nx);
    for(size_t i=0;i<nx;i++) {
      icol[i]=si[i]->deriv(y);
    }
    if (it_set) {
      sm_interp_vec six(*bimp1,*bimp2,nx,*xfun,icol);
      result=six.deriv(x);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.deriv(x);
    }
  }
  return result;
}

double twod_intp::deriv_y(double x, double y) {
  if (data_set==false) {
    O2SCL_ERR("Data not set in twod_intp::interp().",gsl_efailed);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(size_t i=0;i<ny;i++) {
      icol[i]=si[i]->interp(x);
    }
    if (it_set) {
      sm_interp_vec siy(*bimp1,*bimp2,ny,*yfun,icol);
      result=siy.deriv(y);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.deriv(y);
    }
  } else {
    ovector icol(nx);
    for(size_t i=0;i<nx;i++) {
      icol[i]=si[i]->deriv(y);
    }
    if (it_set) {
      sm_interp_vec six(*bimp1,*bimp2,nx,*xfun,icol);
      result=six.interp(x);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.interp(x);
    }
  }
  return result;
}

double twod_intp::deriv2_x(double x, double y) {
  if (data_set==false) {
    O2SCL_ERR("Data not set in twod_intp::interp().",gsl_efailed);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(size_t i=0;i<ny;i++) {
      icol[i]=si[i]->deriv2(x);
    }
    if (it_set) {
      sm_interp_vec siy(*bimp1,*bimp2,ny,*yfun,icol);
      result=siy.interp(y);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.interp(y);
    }
  } else {
    ovector icol(nx);
    for(size_t i=0;i<nx;i++) {
      icol[i]=si[i]->interp(y);
    }
    if (it_set) {
      sm_interp_vec six(*bimp1,*bimp2,nx,*xfun,icol);
      result=six.deriv2(x);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.deriv2(x);
    }
  }
  return result;
}

double twod_intp::deriv2_y(double x, double y) {
  if (data_set==false) {
    O2SCL_ERR("Data not set in twod_intp::interp().",gsl_efailed);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(size_t i=0;i<ny;i++) {
      icol[i]=si[i]->interp(x);
    }
    if (it_set) {
      sm_interp_vec siy(*bimp1,*bimp2,ny,*yfun,icol);
      result=siy.deriv2(y);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.deriv2(y);
    }
  } else {
    ovector icol(nx);
    for(size_t i=0;i<nx;i++) {
      icol[i]=si[i]->deriv2(y);
    }
    if (it_set) {
      sm_interp_vec six(*bimp1,*bimp2,nx,*xfun,icol);
      result=six.interp(x);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.interp(x);
    }
  }
  return result;
}

double twod_intp::integ_x(double x0, double x1, double y) {
  if (data_set==false) {
    O2SCL_ERR("Data not set in twod_intp::interp().",gsl_efailed);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(size_t i=0;i<ny;i++) {
      icol[i]=si[i]->integ(x0,x1);
    }
    if (it_set) {
      sm_interp_vec siy(*bimp1,*bimp2,ny,*yfun,icol);
      result=siy.interp(y);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.interp(y);
    }
  } else {
    ovector icol(nx);
    for(size_t i=0;i<nx;i++) {
      icol[i]=si[i]->interp(y);
    }
    if (it_set) {
      sm_interp_vec six(*bimp1,*bimp2,nx,*xfun,icol);
      result=six.integ(x0,x1);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.integ(x0,x1);
    }
  }
  return result;
}

double twod_intp::integ_y(double x, double y0, double y1) {
  if (data_set==false) {
    O2SCL_ERR("Data not set in twod_intp::interp().",gsl_efailed);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(size_t i=0;i<ny;i++) {
      icol[i]=si[i]->interp(x);
    }
    if (it_set) {
      sm_interp_vec siy(*bimp1,*bimp2,ny,*yfun,icol);
      result=siy.integ(y0,y1);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.integ(y0,y1);
    }
  } else {
    ovector icol(nx);
    for(size_t i=0;i<nx;i++) {
      icol[i]=si[i]->integ(y0,y1);
    }
    if (it_set) {
      sm_interp_vec six(*bimp1,*bimp2,nx,*xfun,icol);
      result=six.interp(x);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.interp(x);
    }
  }
  return result;
}

int twod_intp::unset_data() {
  free_data();
  return 0;
}
