/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_FERMION_H
#define O2SCL_FERMION_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>

// For gsl_sf_fermi_dirac_int()
#include <gsl/gsl_specfunc.h>

#include <o2scl/constants.h>
#include <o2scl/funct.h>
#include <o2scl/root.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/part.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Fermion class

      This is a base class for the computation of fermionic statistics
      at zero temperature. The more general case of finite temperature
      is taken care of by \ref fermion_T objects. The primary
      functions are calc_mu_zerot() and calc_density_zerot() which
      compute all the thermodynamic quantities as a function of the
      chemical potential, or the density, respectively.
      
      This class also adds two member data variables, \ref kf and \ref
      del, for the Fermi momentum and the gap. 

      \future Use hypot() and other more accurate functions for the
      analytic expressions for the zero temperature integrals. [Progress
      has been made, but there are probably other functions which may
      break down for small but finite masses and temperatures]

  */
  class fermion : virtual public part {

  public:

    /// Fermi momentum
    double kf;
    /// Gap
    double del;

    /// Create a fermion with mass \c mass and degeneracy \c dof.
    fermion(double mass=0, double dof=0);

    virtual ~fermion() {
    }

    /// \name Zero-temperature fermions 
    //@{
    /** 
	\brief Calculate the Fermi momentum from the density
	
	Uses the relation \f$ k_F = ( 6 \pi^2 n /g )^{1/3} \f$
    */
    int kf_from_density(); 

    /** 
	\brief Energy density at T=0 from \ref kf and \ref ms

	Calculates the integral 
	\f[
	\varepsilon = \frac{g}{2 \pi^2} \int_0^{k_F} k^2 
	\sqrt{k^2+m^{* 2}} d k
	\f]
    */
    int energy_density_zerot(); 

    /** 
	\brief Pressure at T=0 from \ref kf and \ref ms

	Calculates the integral 
	\f[
	P=\frac{g}{6 \pi^2} \int_0^{k_F} \frac{k^4}{\sqrt{k^2+m^{* 2}}} d k
	\f]
    */
    int pressure_zerot(); 

    /** 
	\brief Zero temperature fermions from nu and ms

	This function always returns \c gsl_success.
     */
    virtual int calc_mu_zerot();

    /** 
	\brief Zero temperature fermions from n and ms

	This function always returns \c gsl_success.
    */
    virtual int calc_density_zerot();
    //@}

    /// Return string denoting type ("fermion")
    virtual const char *type() { return "fermion"; }

  };
  
  /** \brief Fermion with finite-temperature thermodynamics
      [abstract base]

      This is an abstract base for the computation of
      finite-temperature fermionic statistics. Different children
      (e.g. \ref eff_fermion and \ref rel_fermion) use different
      techniques to computing the momentum integrations.

      Because massless fermions at finite temperature are much
      simpler, there are separate member functions included in this
      class to handle them. The functions massless_calc_density() and
      massless_calc_mu() compute the thermodynamics of massless
      fermions at finite temperature given the density or the chemical
      potentials. The functions massless_pair_density() and
      massless_pair_mu() perform the same task, but automatically
      include antiparticles.

      The function massless_calc_density() uses a \ref root object to
      solve for the chemical potential as a function of the density.
      The default is an object of type cern_mroot_root. The function
      massless_pair_density() does not need to use the \ref root
      object because of the simplification afforded by the inclusion
      of antiparticles.

      \future Create a Chebyshev approximation for inverting the 
      the Fermi functions for massless_calc_density() functions?
      
      \htmlonly
      This Mathematica notebook contains the derivations of related
      series expansions and some algebra for the massless_pair()
      functions.
      <a href="../extras/fermion.nb">
      fermion.nb</a>, and
      <a href="../extras/fermion.pdf">
      fermion.pdf</a>.
      \endhtmlonly
      \latexonly
      This Mathematica notebook contains the derivations of related
      series expansions and some algebra for the massless\_pair()
      functions.
      \begin{verbatim}
      doc/o2scl/extras/fermion.nb
      doc/o2scl/extras/fermion.pdf
      \end{verbatim}
      \endlatexonly
  */
  class fermion_T : public fermion {

  public:

    /// Create a fermion with mass \c mass and degeneracy \c dof.
    fermion_T(double mass=0, double dof=0);

    virtual ~fermion_T() {
    }

    /** 
	\brief Calculate properties as function of chemical potential
     */
    virtual int calc_mu(const double temper)=0;

    /** \brief Calculate properties as function of density
     */
    virtual int calc_density(const double temper)=0;

    /** \brief Calculate properties with antiparticles as function of
	chemical potential
    */
    virtual int pair_mu(const double temper)=0;

    /** \brief Calculate properties with antiparticles as function of
	density
    */
    virtual int pair_density(const double temper)=0;

#ifdef O2SCL_NEVER_DEFINED
    /** 
	\brief Degenerate expansion for specific heat
	
	This is a temporary location and is also unchecked.
    */
    double deg_specific_heat(double T) {
      double ret;
      if (non_interacting) {
	nu=mu;
	ms=m;
      }
      double sqd=nu*nu-ms*ms;
      ret=o2scl_const::pi2*T*nu/sqd*
	(1.0-o2scl_const::pi2*T*T*
	 (5.0*pow(ms,4.0)+4.0*ms*ms*nu*nu+14.0*pow(nu,4.0))/
	 15.0/nu/nu/sqd/sqd);
      return ret;
    }
#endif

    /// \name Massless fermions
    //@{
    /// Finite temperature massless fermions
    virtual int massless_calc_mu(const double temper);

    /// Finite temperature massless fermions
    virtual int massless_calc_density(const double temper);

    /** 
	\brief Finite temperature massless fermions and antifermions 
    */
    int massless_pair_mu(const double temper);

    /** 
	\brief Finite temperature massless fermions and antifermions 

	In the cases \f$ n^3 >> T \f$ and \f$ T >> n^3 \f$ ,
	expansions are used instead of the exact formulas to avoid
	loss of precision.

	In particular, using the parameter
	\f[
	\alpha = \frac{g^2 \pi^2 T^6}{243 n^2}
	\f]
	and defining the expression
	\f[
	\mathrm{cbt} = \alpha^{-1/6} \left( -1 + \sqrt{1+\alpha}\right)^{1/3}
	\f]
	we can write the chemical potential as
	\f[
	\mu = \frac{\pi T}{\sqrt{3}} \left(\frac{1}{\mathrm{cbt}} -
	\mathrm{cbt} \right)
	\f]
	
	These expressions, however, do not work well when \f$ \alpha
	\f$ is very large or very small, so series expansions are
	used whenever \f$ \alpha > 10^{4} \f$ or 
	\f$ \alpha < 3 \times 10^{-4} \f$. For large \f$ \alpha \f$, 
	\f[
	\left(\frac{1}{\mathrm{cbt}} -
	\mathrm{cbt} \right) \approx
	\frac{2^{1/3}}{\alpha^{1/6}} - 
	\frac{\alpha^{1/6}}{2^{1/3}} +
	\frac{\alpha^{5/6}}{6~2^{2/3}} +
	\frac{\alpha^{7/6}}{12~2^{1/3}} -
	\frac{\alpha^{11/6}}{18~2^{2/3}} -
	\frac{5 \alpha^{13/6}}{144~2^{1/3}} +
	\frac{77 \alpha^{17/6}}{2592~2^{2/3}}
	\f]
	and for small \f$ \alpha \f$, 
	\f[
	\left(\frac{1}{\mathrm{cbt}} -
	\mathrm{cbt} \right) \approx
	\frac{2}{3} \sqrt{\frac{1}{\alpha}} - 
	\frac{8}{81} \left(\frac{1}{\alpha}\right)^{3/2} +
	\frac{32}{729} \left(\frac{1}{\alpha}\right)^{5/2}
	\f]

	This approach works to within about 1 \part in \f$ 10^{12} \f$,
	and is tested in <tt>fermion_ts.cpp</tt>.
	
	\future This could be improved by including more terms
	in the expansions.
      
    */
    int massless_pair_density(const double temper);
    //@}
    
    /** \brief Set the solver for use in massless_calc_density() */ 
    int set_massless_root(root<const double, funct<const double> > &rp) {
      massless_root=&rp;
      return 0;
    }

    /** 
	\brief The default solver for massless_calc_density()
	
	We default to cern_mroot_root here since we don't have
	a bracket or a derivative.
    */
    cern_mroot_root<const double, funct<const double> > def_massless_root;

    /// Return string denoting type ("fermion_T")
    virtual const char *type() { return "fermion_T"; }

#ifndef DOXYGENP

  protected:
    
    /// A pointer to the solver for massless fermions
    root<const double, funct<const double> > *massless_root;
    
    /// The function to solve for massless fermions
    int massless_fun(double x, double &y, const double &temper);

#endif
    
  };

#ifndef DOXYGENP
}
#endif

#endif
