/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2009, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_REL_BOSON_H
#define O2SCL_REL_BOSON_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/root.h>
#include <o2scl/mroot.h>
#include <o2scl/inte.h>
#include <o2scl/boson.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/gsl_inte_qag.h>
#include <o2scl/gsl_inte_qagiu.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Equation of state for a relativistic boson

      \todo Testing not completely finished.
  */
  class rel_boson : public boson {

  public:

    /// Create a boson with mass \c m and degeneracy \c g
    rel_boson(double m=0.0, double g=0.0);

    virtual ~rel_boson();

    /// The error value from mroot
    int mroot_err;

    /// The error value from inte
    int inte_err;
    
    /**
       \brief Calculate properties as function of chemical potential
    */
    virtual int calc_mu(const double temper);
    
    /** 
	\brief Calculate properties as function of density
    */
    virtual int calc_density(const double temper);
    
    /** \brief Calculate properties with antiparticles as function of
        chemical potential
    */
    virtual int pair_mu(const double temper);

    /** \brief Calculate properties with antiparticles as function of
	density
    */
    virtual int pair_density(const double temper) {
      O2SCL_ERR_RET("Function rel_boson::pair_density() unimplemented.",
		    gsl_eunimpl);
    }

    /// Calculate effective chemical potential from density
    virtual int nu_from_n(const double temper);

    /// Set inte object
    int set_inte(inte<const double,funct<const double> > &l_nit, 
		 inte<const double,funct<const double> > &l_dit);

    /** \brief Set the solver for use in calculating the chemical
	potential from the density */
    int set_density_root(root<const double,funct<const double> > &rp) {
      density_root=&rp;
      return 0;
    }

    /// The default solver for calc_density().
    cern_mroot_root<const double,funct<const double> > def_density_root;

    /// Default nondegenerate integrator
    gsl_inte_qagiu<const double,funct<const double> > def_nit;

    /// Default degenerate integrator
    gsl_inte_qag<const double,funct<const double> > def_dit;

    /// Return string denoting type ("rel_boson")
    virtual const char *type() { return "rel_boson"; }

  protected:

#ifndef DOXYGENP

    /// The non-degenerate integrator
    inte<const double,funct<const double> > *nit;
    /// The degenerate integrator
    inte<const double,funct<const double> > *dit;
    /// The solver for calc_density()
    root<const double,funct<const double> > *density_root;
    
    /// Non-degenerate density integral
    double density_fun(double u, const double &temper);
    /// Non-degenerate energy density integral
    double energy_fun(double u, const double &temper);
    /// Non-degenerate entropy integral
    double entropy_fun(double u, const double &temper);
    /// Degenerate density integral
    double deg_density_fun(double u, const double &temper);
    /// Degenerate energy density integral
    double deg_energy_fun(double u, const double &temper);
    /// Degenerate entropy integral
    double deg_entropy_fun(double u, const double &temper);
    /// Solve for the density in calc_density()
    int solve_fun(double x, double &yy, const double &temper);

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
